"""Module with iterable-related utility functions."""

from collections.abc import Sequence
from typing import TypeVar

__all__ = ["is_ordered_subset"]

T = TypeVar("T")


def is_ordered_subset(one: Sequence[T], other: Sequence[T]) -> bool:
    """Return True if the first iterable is a subset of the second iterable.

    This method is specifically optimized for ordered iterables to use return as early as possible for non-subsets.

    Parameters
    ----------
    one : Sequence
        The first iterable.
    other : Sequence
        The second iterable.

    Returns
    -------
    bool
        True if the first iterable is a subset of the second iterable, False otherwise.

    Examples
    --------
    >>> is_ordered_subset((2,), (1, 2))
    True
    >>> is_ordered_subset((1, 2), (1, 2, 3))
    True
    >>> is_ordered_subset((1, 2, 3), (1, 2, 3))
    True
    >>> is_ordered_subset((1, 2, 3), (1, 2))
    False
    >>> is_ordered_subset((1, 2, 3), (1, 2, 4))
    False
    """
    index = 0
    for item in one:
        while (
            index < len(other)
            and isinstance(item, type(other[index]))
            and other[index] < item
        ):
            index += 1
        if index >= len(other) or other[index] != item:
            return False
    return True



"""Utilities on structures.

The first function, sparse_array_to_neighborhood_list, takes a sparse array
representing a higher order structure between two sets of
cells, S and T, and converts it to a neighborhood list. This
is a list of tuples such that each tuple has the form (s, t),
where s and t are indices representing cells in S and T.
The function also allows for optional dictionaries that map
 the indices in S and T to other values.

The second function, neighborhood_list_to_neighborhood_dict,
takes a neighborhood list and converts it to a neighborhood dictionary.
This dictionary maps each cell i in S to a list of cells j that are in
the neighborhood of i.

The third function, sparse_array_to_neighborhood_dict,
combines the first two functions and takes a sparse array as
input and returns a neighborhood dictionary. It first converts
the sparse array to a neighborhood list, and then converts the
neighborhood list to a neighborhood dictionary. Like the first
function, it also allows for optional dictionaries that map the
indices in S and T to other values.
"""

from collections import OrderedDict, defaultdict
from collections.abc import Iterable
from operator import itemgetter

import numpy as np
from scipy.sparse import csr_matrix

__all__ = [
    "compute_set_incidence",
    "incidence_to_adjacency",
    "neighborhood_list_to_neighborhood_dict",
    "sparse_array_to_neighborhood_dict",
    "sparse_array_to_neighborhood_list",
]


def sparse_array_to_neighborhood_list(
    sparse_array, src_dict=None, dst_dict=None
) -> zip:
    """Convert sparse array to neighborhood list for arbitrary higher order structures.

    Notes
    -----
    neighborhood list is a list of tuples such that each tuple has the form (s,t), s and t
        are indices representing a cell in a higher order structure.
        this structure can be converted to a matrix of size |S|X|T| where |S| is
        the size of the source cells and |T| is the size of the target cells.

    Parameters
    ----------
    sparse_array : sparse array
        Sparse array representing the higher order structure between S and T cells.
    src_dict : dict
        Dictionary mapping the indices in S to other values.
    dst_dict : dict
        Dictionary mapping the indices in T to other values.

    Returns
    -------
    tuple(tuple[int, int])
        Zip object of the form (s,t) where s and t are indices representing a cell in a higher order structure.
    """
    src_idx, dst_idx = sparse_array.nonzero()

    if src_dict is None and dst_dict is None:
        return zip(dst_idx, src_idx, strict=True)

    if src_dict is not None and dst_dict is not None:
        src_list = [src_dict[i] for i in src_idx]
        dest_list = [dst_dict[i] for i in dst_idx]
        return zip(dest_list, src_list, strict=True)

    raise ValueError("src_dict and dst_dict must be either None or both not None")


def neighborhood_list_to_neighborhood_dict(
    n_list: Iterable[tuple[int, int]], src_dict=None, dst_dict=None
) -> dict[int, list[int]]:
    """Convert neighborhood list to neighborhood dictionary for arbitrary higher order structures.

    Notes
    -----
        for every cell i, neighborhood_dict[i] is describe all cells j that are in the neighborhood to j.

    Parameters
    ----------
        n_list : list[tuple[int, int]]
            Neighborhood list.
        src_dict : dict
            Dictionary mapping the indices in S to other values.
        dst_dict : dict
            Dictionary mapping the indices in T to other values.

    Returns
    -------
        dict[int, list[int]]
            Neighborhood dictionary.
    """
    neighborhood_dict = defaultdict(list)
    if src_dict is None and dst_dict is None:
        for src_idx, dst_idx in n_list:
            neighborhood_dict[src_idx].append(dst_idx)
        return neighborhood_dict

    if src_dict is not None and dst_dict is not None:
        for src_idx, dst_idx in n_list:
            neighborhood_dict[src_dict[src_idx]].append(dst_dict[dst_idx])
        return neighborhood_dict

    raise ValueError("src_dict and dst_dict must be either None or both not None")


def sparse_array_to_neighborhood_dict(
    sparse_array, src_dict=None, dst_dict=None
) -> dict[int, list[int]]:
    """Convert sparse array to neighborhood dictionary for arbitrary higher order structures.

    Notes
    -----
        neighborhood list is a list of tuples such that each tuple has the form (s,t), s and t
        are indices representing a cell in a higher order structure.
        this structure can be converted to a matrix of size |S|X|T| where |S| is
        the size of the source cells and |T| is the size of the target cells.

    Parameters
    ----------
        sparse_array : sparse array
            Sparse array representing the higher order structure between S and T cells.
        src_dict : dict
            Dictionary mapping the indices in S to other values.
        dst_dict : dict
            Dictionary mapping the indices in T to other values.

    Returns
    -------
        dict[int, list[int]]
            Neighborhood dictionary.
    """
    return neighborhood_list_to_neighborhood_dict(
        sparse_array_to_neighborhood_list(sparse_array, src_dict, dst_dict)
    )


def incidence_to_adjacency(B, s: int = 1, signed: bool = False):
    """Get adjacency matrix from boolean incidence matrix for s-metrics.

    Self loops are not supported.
    The adjacency matrix will define an s-linegraph.

    Parameters
    ----------
    B : scipy.sparse.csr.csr_matrix
        Incidence matrix of 0's and 1's.
    s : int, list, default=1
        Minimum number of edges shared by neighbors with node.
    signed : bool, optional
        If not signed, then we take the absolute values.

    Returns
    -------
    scipy.sparse.csr.csr_matrix
        Return the adjacency matrix from booloean incidence matrix.
    """
    B = csr_matrix(B)
    if not signed:
        B = abs(B)  # make sure the incidence matrix has only positive entries

    A = B.T @ B
    A.setdiag(0)
    A.data[A.data < s] = 0
    return A


def compute_set_incidence(children, uidset, sparse: bool = True, index: bool = False):
    """Compute set-based incidence.

    Parameters
    ----------
    children : list
        List of children.
    uidset : list
        List of uidset.
    sparse : bool, optional
        If True, return a sparse matrix.
    index : bool, optional
        If True, return the index dictionaries.

    Returns
    -------
    ndict : dict
        Index dictionary for children.
    edict : dict
        Index dictionary for uidset.
    MP : scipy.sparse.csr.csr_matrix
        Set-based incidence matrix.
    """
    ndict = dict(zip(children, range(len(children)), strict=True))
    edict = dict(zip(uidset, range(len(uidset)), strict=True))

    ndict = OrderedDict(sorted(ndict.items(), key=itemgetter(1)))
    edict = OrderedDict(sorted(edict.items(), key=itemgetter(1)))

    r_hyperedge_dict = {j: children[j] for j in range(len(children))}
    k_hyperedge_dict = {i: uidset[i] for i in range(len(uidset))}

    r_hyperedge_dict = OrderedDict(sorted(r_hyperedge_dict.items(), key=itemgetter(0)))
    k_hyperedge_dict = OrderedDict(sorted(k_hyperedge_dict.items(), key=itemgetter(0)))

    if len(ndict) != 0:
        if sparse:
            # Create csr sparse matrix
            rows = []
            cols = []
            data = []
            for n in ndict:
                for e in edict:
                    if n <= e:
                        data.append(1)
                        rows.append(ndict[n])
                        cols.append(edict[e])
            MP = csr_matrix(
                (data, (rows, cols)),
                shape=(len(r_hyperedge_dict), len(k_hyperedge_dict)),
            )
        else:
            # Create an np.matrix
            MP = np.zeros((len(children), len(uidset)), dtype=int)
            for e in k_hyperedge_dict:
                for n in r_hyperedge_dict:
                    if r_hyperedge_dict[n] <= k_hyperedge_dict[e]:
                        MP[ndict[r_hyperedge_dict[n]], edict[k_hyperedge_dict[e]]] = 1
        if index:
            return ndict, edict, MP
        return MP

    if index:
        return {}, {}, np.zeros(1)
    return np.zeros(1)



"""Various examples of named graphs represented as complexes."""

from pathlib import Path
from typing import Literal, overload

import networkx as nx
import numpy as np
import requests
from typing_extensions import deprecated

from toponetx.algorithms.spectrum import hodge_laplacian_eigenvectors
from toponetx.classes.cell_complex import CellComplex
from toponetx.classes.simplicial_complex import SimplicialComplex
from toponetx.transform.graph_to_simplicial_complex import graph_to_clique_complex

__all__ = ["coauthorship", "karate_club"]

DIR = Path(__file__).parent


@overload
def karate_club(
    complex_type: Literal["cell"], feat_dim: int = ...
) -> CellComplex:  # numpydoc ignore=GL08
    ...


@overload
def karate_club(
    complex_type: Literal["simplicial"], feat_dim: int = ...
) -> SimplicialComplex:  # numpydoc ignore=GL08
    ...


def karate_club(
    complex_type: Literal["cell", "simplicial"], feat_dim: int = 2
) -> CellComplex | SimplicialComplex:
    """Load the karate club as featured cell/simplicial complex.

    Parameters
    ----------
    complex_type : {'simplicial','cell'}
        The type of complex to load.
    feat_dim : int, default=2
        The number of eigenvectors to be attached to the simplices/cells
        of the output complex.

    Returns
    -------
    SimplicialComplex or CellComplex
        When input is "simplicial":
        A SimplicialComplex obtained from karate club graph by lifting the graph to its
        clique complex. The simplicial complex comes with the following features:

        - "node_feat": its value is the first feat_dim Hodge Laplacian eigenvectors attached to nodes.
        - "edge_feat": its value is the first feat_dim Hodge Laplacian eigenvectors attached to edges.
        - "face_feat": its value is the first feat_dim Hodge Laplacian eigenvectors attached to faces.
        - "tetrahedron_feat": the first feat_dim Hodge Laplacian eigenvectors attached to tetrahedron.

        When input is "cell":
        A CellComplex obtained from karate club by lifting the graph to a cell obtained
        obtained from the graph by adding the independent homology cycles in the graph.
        The cell complex comes with the following features:

        - "node_feat": its value is the first feat_dim Hodge Laplacian eigenvectors
          attached to nodes.
        - "edge_feat": its value is the first feat_dim Hodge Laplacian eigenvectors
          attached to edges.
        - "cell_feat": its value is the first feat_dim Hodge Laplacian eigenvectors
          attached to cells.

    Raises
    ------
    ValueError
        If complex_type is not one of the supported values.

    Notes
    -----
    A featured simplicial complex is returned as the clique complex of the graph.
    A featured cell complex is returned as the cell complex obtained by adding the
    independent cycles of graph.
    """
    if complex_type == "simplicial":
        g = nx.karate_club_graph()
        sc = graph_to_clique_complex(g)

        _, nodes_feat = hodge_laplacian_eigenvectors(
            sc.hodge_laplacian_matrix(0), feat_dim
        )
        _, edges_feat = hodge_laplacian_eigenvectors(
            sc.hodge_laplacian_matrix(1), feat_dim
        )
        _, faces_feat = hodge_laplacian_eigenvectors(
            sc.hodge_laplacian_matrix(2), feat_dim
        )
        _, tetrahedron_feat = hodge_laplacian_eigenvectors(
            sc.hodge_laplacian_matrix(3), feat_dim
        )

        sc.set_simplex_attributes(
            dict(zip(sc.skeleton(0), np.array(nodes_feat), strict=True)),
            name="node_feat",
        )
        sc.set_simplex_attributes(
            dict(zip(sc.skeleton(1), np.array(edges_feat), strict=True)),
            name="edge_feat",
        )
        sc.set_simplex_attributes(
            dict(zip(sc.skeleton(2), np.array(faces_feat), strict=True)),
            name="face_feat",
        )
        sc.set_simplex_attributes(
            dict(zip(sc.skeleton(3), np.array(tetrahedron_feat), strict=True)),
            name="tetrahedron_feat",
        )

        return sc

    if complex_type == "cell":
        g = nx.karate_club_graph()
        cycles = nx.cycle_basis(g)
        cx = CellComplex(g)
        cx.add_cells_from(cycles, rank=2)

        _, nodes_feat = hodge_laplacian_eigenvectors(
            cx.hodge_laplacian_matrix(0), feat_dim
        )
        _, edges_feat = hodge_laplacian_eigenvectors(
            cx.hodge_laplacian_matrix(1), feat_dim
        )
        _, cells_feat = hodge_laplacian_eigenvectors(
            cx.hodge_laplacian_matrix(2), feat_dim
        )

        cx.set_cell_attributes(
            dict(zip(cx.skeleton(0), np.array(nodes_feat), strict=True)),
            name="node_feat",
            rank=0,
        )
        cx.set_cell_attributes(
            dict(zip(cx.skeleton(1), np.array(edges_feat), strict=True)),
            name="edge_feat",
            rank=1,
        )
        cx.set_cell_attributes(
            dict(zip(cx.skeleton(2), np.array(cells_feat), strict=True)),
            name="cell_feat",
            rank=2,
        )
        return cx

    raise ValueError(f"complex_type must be 'simplicial' or 'cell' got {complex_type}")


@deprecated(
    "coauthorship is deprecated since version 0.3.0 and will be removed in a future version. "
    "The dataset is available in AHORN under 'semantic-scholar-coauth-sample'."
)
def coauthorship() -> SimplicialComplex:
    """Load the coauthorship network as a simplicial complex.

    The coauthorship network is a simplicial complex where a paper with k authors is
    represented by a (k-1)-simplex.
    The dataset is pre-processed as in [1]_. From the
    Semantic Scholar Open Research Corpus 80 papers with number of citations between 5
    and 10 were sampled.

    The papers constitute simplices in the complex, which is completed with
    sub-simplices (seen as collaborations between subsets of authors) to form a
    simplicial complex.
    An attribute named *citations* is added to each simplex, corresponding to the sum
    of citations of all papers on which the authors represented by the simplex
    collaborated. The resulting simplicial complex is of dimension 10 and contains
    24552 simplices in total. See [1]_ for a more detailed description of the dataset.

    Returns
    -------
    SimplicialComplex
        The simplicial complex comes with the attribute *citations*, the number of
        citations attributed to the given collaborations of k authors.

    References
    ----------
    .. [1] Stefania Ebli, Michael Defferrard and Gard Spreemann. Simplicial Neural
        Networks. Topological Data Analysis and Beyond workshop at NeurIPS.
        https://arxiv.org/abs/2010.03633
    """
    dataset_file = DIR / "coauthorship.npy"

    if not dataset_file.exists():
        r = requests.get(
            "https://github.com/pyt-team/topological-datasets/raw/main/resources/coauthorship.npy",
            timeout=10,
        )
        with dataset_file.open("wb") as f:
            f.write(r.content)

    coauthorship = np.load(dataset_file, allow_pickle=True)

    simplices = []
    for dim in range(len(coauthorship) - 1, -1, -1):
        simplices += [list(el) for el in coauthorship[dim]]

    SC = SimplicialComplex(simplices)

    for i in range(len(coauthorship)):
        dic = {tuple(sorted(k)): v for k, v in coauthorship[i].items()}
        SC.set_simplex_attributes(dic, name="citations")

    return SC



"""
Implementation of a simplex trie datastructure for simplicial complexes as presented in [1]_.

This module is intended for internal use by the `SimplicialComplex` class only. Any direct interactions with this
module or its classes may break at any time. In particular, this also means that the `SimplicialComplex` class must not
leak any object references to the trie or its nodes.

Some implementation details:
- Inside this module, simplices are represented as ordered sequences with unique elements. It is expected that all
  inputs from outside are already pre-processed and ordered accordingly. This is not checked and the behavior is
  undefined if this is not the case.

References
----------
.. [1] Jean-Daniel Boissonnat and Clément Maria. The Simplex Tree: An Efficient Data Structure for General Simplicial
       Complexes. Algorithmica, pages 1-22, 2014
"""

from collections.abc import Generator, Hashable, Iterable, Iterator, Sequence
from typing import Any

from toponetx.classes.simplex import Simplex
from toponetx.utils.iterable import is_ordered_subset

__all__ = ["SimplexNode", "SimplexTrie"]


class SimplexNode[ElementType: Hashable]:
    """Node in a simplex tree.

    Parameters
    ----------
    label : ElementType or None
        The label of the node. May only be `None` for the root node.
    parent : SimplexNode, optional
        The parent node of this node. If `None`, this node is the root node.
    """

    label: ElementType | None
    elements: tuple[ElementType, ...]
    attributes: dict[Hashable, Any]

    depth: int
    parent: "SimplexNode | None"
    children: dict[ElementType, "SimplexNode[ElementType]"]

    def __init__(
        self,
        label: ElementType | None,
        parent: "SimplexNode[ElementType] | None" = None,
    ) -> None:
        """Node in a simplex tree.

        Parameters
        ----------
        label : ElementType or None
            The label of the node. May only be `None` for the root node.
        parent : SimplexNode, optional
            The parent node of this node. If `None`, this node is the root node.
        """
        self.label = label
        self.attributes = {}

        self.children = {}

        self.parent = parent
        if parent is not None:
            parent.children[label] = self
            self.elements = (*parent.elements, label)
            self.depth = parent.depth + 1
        else:
            if label is not None:
                raise ValueError("Root node must have label `None`.")
            self.elements = ()
            self.depth = 0

    def __len__(self) -> int:
        """Return the number of elements in this trie node.

        Returns
        -------
        int
            Number of elements in this trie node.
        """
        return len(self.elements)

    def __repr__(self) -> str:
        """Return a string representation of this trie node.

        Returns
        -------
        str
            A string representation of this trie node.
        """
        return f"SimplexNode({self.label}, {self.parent!r})"

    @property
    def simplex(self) -> Simplex[ElementType] | None:
        """Return a `Simplex` object representing this node.

        Returns
        -------
        Simplex or None
            A `Simplex` object representing this node, or `None` if this node is the root node.
        """
        if self.label is None:
            return None
        simplex = Simplex(self.elements, construct_tree=False)
        simplex._attributes = self.attributes
        return simplex

    def iter_all(self) -> Generator["SimplexNode[ElementType]", None, None]:
        """Iterate over all nodes in the subtree rooted at this node.

        Ordering is according to breadth-first search, i.e., simplices are yielded in increasing order of dimension and
        increasing order of labels within each dimension.

        Yields
        ------
        SimplexNode
        """
        queue = [self]
        while queue:
            node = queue.pop(0)
            if node.label is not None:
                # skip root node
                yield node
            queue += [node.children[label] for label in sorted(node.children.keys())]


class SimplexTrie[ElementType: Hashable]:
    """
    Simplex tree data structure as presented in [1]_.

    This class is intended for internal use by the `SimplicialComplex` class only. Any
    direct interactions with this class may break at any time.

    References
    ----------
    .. [1] Jean-Daniel Boissonnat and Clément Maria. The Simplex Tree: An Efficient
           Data Structure for General Simplicial Complexes. Algorithmica, pages 1-22, 2014
    """

    root: SimplexNode[ElementType]
    label_lists: dict[int, dict[ElementType, list[SimplexNode[ElementType]]]]
    shape: list[int]

    def __init__(self) -> None:
        """Simplex tree data structure as presented in [1]_.

        This class is intended for internal use by the `SimplicialComplex` class only.
        Any direct interactions with this class may break at any time.

        References
        ----------
        .. [1] Jean-Daniel Boissonnat and Clément Maria. The Simplex Tree: An Efficient
               Data Structure for General Simplicial Complexes. Algorithmica, pages 1-22, 2014
        """
        self.root = SimplexNode(None)
        self.label_lists = {}
        self.shape = []

    def __len__(self) -> int:
        """Return the number of simplices in the trie.

        Returns
        -------
        int
            Number of simplices in the trie.

        Examples
        --------
        >>> trie = SimplexTrie()
        >>> trie.insert((1, 2, 3))
        >>> len(trie)
        7
        """
        return sum(self.shape)

    def __contains__(self, item: Iterable[ElementType]) -> bool:
        """Check if a simplex is contained in this trie.

        Parameters
        ----------
        item : Iterable of ElementType
            The simplex to check for. Must be ordered and contain unique elements.

        Returns
        -------
        bool
            Whether the given simplex is contained in this trie.

        Examples
        --------
        >>> trie = SimplexTrie()
        >>> trie.insert((1, 2, 3))
        >>> (1, 2, 3) in trie
        True
        >>> (1, 2, 4) in trie
        False
        """
        return self.find(item) is not None

    def __getitem__(self, item: Iterable[ElementType]) -> SimplexNode[ElementType]:
        """Return the simplex node for a given simplex.

        Parameters
        ----------
        item : Iterable of ElementType
            The simplex to return the node for. Must be ordered and contain only unique
            elements.

        Returns
        -------
        SimplexNode
            The trie node that represents the given simplex.

        Raises
        ------
        KeyError
            If the given simplex does not exist in this trie.
        """
        node = self.find(item)
        if node is None:
            raise KeyError(f"Simplex {item} not found in trie.")
        return node

    def __iter__(self) -> Iterator[SimplexNode[ElementType]]:
        """Iterate over all simplices in the trie.

        Simplices are ordered by increasing dimension and increasing order of labels within each dimension.

        Yields
        ------
        tuple of ElementType
        """
        yield from self.root.iter_all()

    def insert(self, item: Sequence[ElementType], **kwargs) -> None:
        """Insert a simplex into the trie.

        Any lower-dimensional simplices that do not exist in the trie are also inserted
        to fulfill the simplex property. If the simplex already exists, its properties
        are updated.

        Parameters
        ----------
        item : Sequence of ElementType
            The simplex to insert. Must be ordered and contain only unique elements.
        **kwargs : keyword arguments, optional
            Properties associated with the given simplex.
        """
        self._insert_helper(self.root, item)
        self.find(item).attributes.update(kwargs)

    # def insert_raw(self, simplex: Sequence[ElementType], **kwargs) -> None:
    #     """Insert a simplex into the trie without guaranteeing the simplex property.

    #     Sub-simplices are not guaranteed to be inserted, which may break the simplex
    #     property. This allows for faster insertion of a sequence of simplices without
    #     considering their order. In case not all sub-simplices are inserted manually,
    #     the simplex property must be restored by calling `restore_simplex_property`.

    #     Parameters
    #     ----------
    #     simplex : Sequence of ElementType
    #         The simplex to insert. Must be ordered and contain only unique elements.
    #     **kwargs : keyword arguments, optional
    #         Properties associated with the simplex.

    #     See Also
    #     --------
    #     restore_simplex_property
    #         Function to restore the simplex property after using `insert_raw`.
    #     """
    #     current_node = self.root
    #     for label in simplex:
    #         if label in current_node.children:
    #             current_node = current_node.children[label]
    #         else:
    #             current_node = self._insert_child(current_node, label)
    #     current_node.attributes.update(kwargs)

    def _insert_helper(
        self, subtree: SimplexNode[ElementType], items: Sequence[ElementType]
    ) -> None:
        """Insert a simplex node under the given subtree.

        Parameters
        ----------
        subtree : SimplexNode
            The subtree to insert the simplex node under.
        items : Sequence
            The (partial) simplex to insert under the subtree.
        """
        for i, label in enumerate(items):
            if label not in subtree.children:
                self._insert_child(subtree, label)
            self._insert_helper(subtree.children[label], items[i + 1 :])

    def _insert_child(
        self, parent: SimplexNode[ElementType], label: ElementType
    ) -> SimplexNode[ElementType]:
        """Insert a child node with a given label.

        Parameters
        ----------
        parent : SimplexNode
            The parent node.
        label : ElementType
            The label of the child node.

        Returns
        -------
        SimplexNode
            The new child node.
        """
        node = SimplexNode(label, parent=parent)

        # update label lists
        if node.depth not in self.label_lists:
            self.label_lists[node.depth] = {}
        if label in self.label_lists[node.depth]:
            self.label_lists[node.depth][label].append(node)
        else:
            self.label_lists[node.depth][label] = [node]

        # update shape property
        if node.depth > len(self.shape):
            self.shape += [0]
        self.shape[node.depth - 1] += 1

        return node

    # def restore_simplex_property(self) -> None:
    #     """Restore the simplex property after using `insert_raw`."""
    #     all_simplices = set()
    #     for node in self.root.iter_all():
    #         if len(node.children) == 0:
    #             for r in range(1, len(node.elements)):
    #                 all_simplices.update(combinations(node.elements, r))

    #     for simplex in all_simplices:
    #         self.insert_raw(simplex)

    def find(self, search: Iterable[ElementType]) -> SimplexNode[ElementType] | None:
        """Find the node in the trie that matches the search.

        Parameters
        ----------
        search : Iterable of ElementType
            The simplex to search for. Must be ordered and contain only unique elements.

        Returns
        -------
        SimplexNode or None
            The node that matches the search, or `None` if no such node exists.
        """
        node = self.root
        for item in search:
            if item not in node.children:
                return None
            node = node.children[item]
        return node

    def cofaces(
        self, simplex: Sequence[ElementType]
    ) -> Generator[SimplexNode[ElementType], None, None]:
        """Return the cofaces of the given simplex.

        No ordering is guaranteed by this method.

        Parameters
        ----------
        simplex : Sequence of ElementType
            The simplex to find the cofaces of. Must be ordered and contain only unique elements.

        Yields
        ------
        SimplexNode
            The cofaces of the given simplex, including the simplex itself.

        Examples
        --------
        >>> trie = SimplexTrie()
        >>> trie.insert((1, 2, 3))
        >>> trie.insert((1, 2, 4))
        >>> sorted(map(lambda node: node.elements, trie.cofaces((1,))))
        [(1,), (1, 2), (1, 2, 3), (1, 2, 4), (1, 3), (1, 4)]
        """
        # Find simplices of the form [*, l_1, *, l_2, ..., *, l_n], i.e. simplices that contain all elements of the
        # given simplex plus some additional elements, but sharing the same largest element. This can be done by the
        # label lists.
        simplex_nodes = self._coface_roots(simplex)

        # Found all simplices of the form [*, l_1, *, l_2, ..., *, l_n] in the simplex trie. All nodes in the subtrees
        # rooted at these nodes are cofaces of the given simplex.
        for simplex_node in simplex_nodes:
            yield from simplex_node.iter_all()

    def _coface_roots(
        self, simplex: Sequence[ElementType]
    ) -> Generator[SimplexNode[ElementType], None, None]:
        """Return the roots of the coface subtrees.

        A coface subtree is a subtree of the trie whose simplices are all cofaces of a
        given simplex.

        Parameters
        ----------
        simplex : Sequence of ElementType
            The simplex to find the cofaces of. Must be ordered and contain only unique
            elements.

        Yields
        ------
        SimplexNode
            The trie nodes that are roots of the coface subtrees.
        """
        for depth in range(len(simplex), len(self.shape) + 1):
            if simplex[-1] not in self.label_lists[depth]:
                continue
            for simplex_node in self.label_lists[depth][simplex[-1]]:
                if is_ordered_subset(simplex, simplex_node.elements):
                    yield simplex_node

    def is_maximal(self, simplex: tuple[ElementType, ...]) -> bool:
        """Return True if the given simplex is maximal.

        A simplex is maximal if it has no cofaces.

        Parameters
        ----------
        simplex : tuple of ElementType or SimplexNode
            The simplex to check. Must be ordered and contain only unique elements.

        Returns
        -------
        bool
            True if the given simplex is maximal, False otherwise.

        Raises
        ------
        ValueError
            If the simplex does not exist in the trie.

        Examples
        --------
        >>> trie = SimplexTrie()
        >>> trie.insert((1, 2, 3))
        >>> trie.is_maximal((1, 2, 3))
        True
        >>> trie.is_maximal((1, 2))
        False
        """
        if simplex not in self:
            raise ValueError(f"Simplex {simplex} does not exist.")

        gen = self._coface_roots(simplex)
        try:
            first_next = next(gen)
            # sometimes the first root is the simplex itself
            if first_next.elements == simplex:
                if len(first_next.children) > 0:
                    return False
                next(gen)
            return False
        except StopIteration:
            return True

    def skeleton(self, rank: int) -> Generator[SimplexNode, None, None]:
        """Return the simplices of the given rank.

        No particular ordering is guaranteed and is dependent on insertion order.

        Parameters
        ----------
        rank : int
            The rank of the simplices to return.

        Yields
        ------
        SimplexNode
            The simplices of the given rank.

        Raises
        ------
        ValueError
            If the given rank is negative or exceeds the maximum rank of the trie.

        Examples
        --------
        >>> trie = SimplexTrie()
        >>> trie.insert((1, 2, 3))
        >>> sorted(map(lambda node: node.elements, trie.skeleton(0)))
        [(1,), (2,), (3,)]
        >>> sorted(map(lambda node: node.elements, trie.skeleton(1)))
        [(1, 2), (1, 3), (2, 3)]
        >>> sorted(map(lambda node: node.elements, trie.skeleton(2)))
        [(1, 2, 3)]
        """
        if rank < 0:
            raise ValueError(f"`rank` must be a positive integer, got {rank}.")
        if rank >= len(self.shape):
            raise ValueError(f"`rank` {rank} exceeds maximum rank {len(self.shape)}.")

        for nodes in self.label_lists[rank + 1].values():
            yield from nodes

    def remove_simplex(self, simplex: Sequence[ElementType]) -> None:
        """Remove the given simplex and all its cofaces from the trie.

        This method ensures that the simplicial property is maintained by removing any simplex that is no longer valid
        after removing the given simplex.

        Parameters
        ----------
        simplex : Sequence of ElementType
            The simplex to remove. Must be ordered and contain only unique elements.

        Raises
        ------
        ValueError
            If the simplex does not exist in the trie.
        """
        # Locate all roots of subtrees containing cofaces of this simplex. They are invalid after removal of the given
        # simplex and thus need to be removed as well.
        coface_roots = self._coface_roots(simplex)
        for subtree_root in coface_roots:
            for node in subtree_root.iter_all():
                self.shape[node.depth - 1] -= 1
                self.label_lists[node.depth][node.label].remove(node)

            # Detach the subtree from the trie. This effectively destroys the subtree as no reference exists anymore,
            # and garbage collection will take care of the rest.
            subtree_root.parent.children.pop(subtree_root.label)

        # After removal of some simplices, the maximal dimension may have decreased. Make sure that there are no empty
        # shape entries at the end of the shape list.
        while len(self.shape) > 0 and self.shape[-1] == 0:
            self.shape.pop()



"""Cell class."""

from collections import Counter, deque
from collections.abc import Collection, Hashable, Iterable, Sequence
from itertools import zip_longest
from typing import Literal, Self

from toponetx.classes.complex import Atom

__all__ = ["Cell"]


class Cell(Atom[tuple[Hashable]]):
    """Class representing a 2D cell.

    A 2D cell is an elementary building block used to build a 2D cell complex, whether regular or non-regular.

    Parameters
    ----------
    elements : iterable of hashable objects
        An iterable that contains hashable objects representing the nodes of the cell. The order of the elements is important
        and defines the cell up to cyclic permutation.
    regular : bool, optional
        A boolean indicating whether the cell satisfies the regularity condition. The default value is True.
        A 2D cell is regular if and only if there is no repetition in the boundary edges that define the cell.
        By default, the cell is assumed to be regular unless otherwise specified. Self-loops are not allowed in the boundary
        of the cell. If a cell violates the cell complex regularity condition, a ValueError is raised.
    **kwargs : keyword arguments, optional
        Attributes belonging to the cell can be added as key-value pairs. Both the key and value must be hashable.

    Notes
    -----
    - A cell is defined as an ordered sequence of nodes (n_1, ..., n_k), where each two
      consecutive nodes (n_i, n_{i+1}) define an edge in the boundary of the cell. Note
      that the last edge (n_k, n_1) is also included in the boundary of the cell. For
      instance, if a Cell is defined as `c = Cell((1, 2, 3))`, then `c.boundary` will
      return `[(1, 2), (2, 3), (3, 1)]`, which consists of three edges.
    - When a cell is created, its boundary is automatically created as a set of edges
      that encircle the cell.

    Examples
    --------
    >>> cell1 = tnx.Cell((1, 2, 3))
    >>> cell2 = tnx.Cell((1, 2, 4, 5), weight=1)
    >>> cell3 = tnx.Cell(("a", "b", "c"))
    >>> # create geometric cell:
    >>> v0 = (0, 0)
    >>> v1 = (1, 0)
    >>> v2 = (1, 1)
    >>> v3 = (0, 1)
    # create the cell with the vertices and edges
    >>> cell = tnx.Cell([v0, v1, v2, v3], type="square")
    >>> cell["type"]
    >>> list(cell.boundary)
    [((0, 0), (1, 0)), ((1, 0), (1, 1)), ((1, 1), (0, 1)),
    ((0, 1), (0, 0))]
    """

    def __init__(self, elements: Collection, regular: bool = True, **kwargs) -> None:
        super().__init__(tuple(elements), **kwargs)

        self._regular = regular
        elements = list(elements)
        self._boundary = list(
            zip_longest(elements, [*elements[1:], elements[0]])
        )  # list of edges defines the boundary of the 2d cell
        if len(elements) <= 1:
            raise ValueError(
                f"cell must contain at least 2 edges, got {len(elements) + 1}"
            )

        if regular:
            _adjdict = {}
            for e in self._boundary:
                if e[0] in _adjdict:
                    raise ValueError(
                        f"Node {e[0]} is repeated multiple times in the input cell. "
                        "Input cell violates the cell complex regularity condition."
                    )
                _adjdict[e[0]] = e[1]
        else:
            for e in self._boundary:
                if e[0] == e[1]:
                    raise ValueError(
                        f"self loops are not permitted, got {(e[0], e[1])} as an edge in the cell's boundary"
                    )

    def clone(self) -> Self:
        """Clone the Cell with all attributes.

        The clone method by default returns an independent shallow copy of the cell and
        attributes. That is, if an attribute is a container, that container is shared
        by the original and the copy. Use Python's `copy.deepcopy` for new containers.

        Returns
        -------
        Cell
            A copy of this cell.
        """
        return self.__class__(self.elements, self._regular, **self._attributes)

    @staticmethod
    def is_valid_cell(elements: Sequence, regular: bool = False) -> bool:
        """Check if a 2D cell defined by a list of elements is valid.

        Parameters
        ----------
        elements : Sequence
            List of elements defining the cell.
        regular : bool, default=False
            Indicates if the cell is regular.

        Returns
        -------
        bool
            True if the cell is valid, False otherwise.
        """
        _boundary = list(
            zip_longest(elements, [*elements[1:], elements[0]])
        )  # list of edges define the boundary of the 2d cell
        if len(elements) <= 1:
            return False

        if regular:
            _adjdict = {}
            for e in _boundary:
                if e[0] in _adjdict:
                    return False

                _adjdict[e[0]] = e[1]
        else:
            for e in _boundary:
                if e[0] == e[1]:
                    return False
        return True

    @property
    def is_regular(self) -> bool:
        """Check if a cell is regular.

        Returns
        -------
        bool
            True if the Cell is regular, and False otherwise.
        """
        if self._regular:  # condition enforced
            return True

        _adjdict = {}
        for e in self._boundary:
            if e[0] in _adjdict:
                return False
            _adjdict[e[0]] = e[1]

        return True

    def sign(self, edge) -> Literal[-1, 1]:
        """Compute the sign of the given edge with respect to this cell.

        This takes an edge as input and computes the sign of the edge with respect to the cell.

        If the edge is in the boundary of the cell, then the sign is 1 if the edge is in the
        counterclockwise direction around the cell and -1 if it is in the clockwise direction.

        If the edge is not in the boundary of the cell, a KeyError is raised.

        Parameters
        ----------
        edge : Iterable
            The edge whose sign should be computed.

        Returns
        -------
        {1, -1}
            1: if the edge is in the counterclockwise direction around the cell.
            -1: if the edge is in the clockwise direction around the cell.

        Raises
        ------
        KeyError
            If the input edge is not in the boundary of the cell.
        ValueError
            If the input edge is not valid.
        TypeError
            If the input edge is not iterable.
        """
        if not isinstance(edge, Iterable):
            raise TypeError(f"The input {edge} must be iterable.")

        edge = tuple(edge)
        if len(edge) == 2:
            if tuple(edge) in self.boundary:
                return 1
            if tuple(edge)[::-1] in self.boundary:
                return -1
            raise KeyError(f"Ihe input {edge} is not in the boundary of the cell.")

        raise ValueError(f"The input {edge} is not a valid edge.")

    def __repr__(self) -> str:
        """Return string representation of regular cell.

        Returns
        -------
        str
            The string representation of the regular cell.
        """
        return f"Cell({self.elements})"

    @property
    def boundary(self):
        """Boundary.

        A 2d cell is characterized by its boundary edges.

        Returns
        -------
        Iterable[tuple[int, int]]
            Iterator of tuple representing boundary edges given in cyclic order.

        Examples
        --------
        For a cell [1,2,3,4,5], the boundary is the edges that define that cell :
            (1,2), (2,3), (3,4), (4,5) and (5,1).
        """
        return iter(self._boundary)

    def reverse(self):
        """Reverse the sequence of nodes that defines the cell.

        Returns
        -------
        Cell
            New cell with the new reversed elements.
        """
        return Cell(
            self.elements[::-1],
            regular=self._regular,
            **self._attributes,
        )

    def is_homotopic_to(self, cell) -> bool:
        """Check if self is homotopic to input cell.

        Parameters
        ----------
        cell : tuple, list or Cell
            The cell to check for homotopy.

        Returns
        -------
        bool
            Return True is self is homotopic to input cell and False otherwise.
        """
        return Cell._are_homotopic(self, cell) or Cell._are_homotopic(
            self.reverse(), cell
        )

    @staticmethod
    def _are_homotopic(cell1, cell) -> bool:
        """Check if cell1 is homotopic to input cell.

        Parameters
        ----------
        cell1 : Cell
            First cell to check for homotopy.
        cell : tuple | list | Cell
            Second cell to check for homotopy.

        Returns
        -------
        bool
            Returns True is self is homotopic to input cell and False otherwise.

        Notes
        -----
        in a 2d-cell complex, two 2d-cells are homotopic iff one
            of them can be obtaine from the other by a cylic rotation
            of the boundary verties defining the 2d cells.
        """
        if isinstance(cell, tuple | list):
            seq = cell
        elif isinstance(cell, Cell):
            seq = cell.elements
        else:
            raise TypeError(
                "fInput {cell} must be a tuple/list of nodes defining a cell or Cell"
            )

        if not isinstance(cell1, Cell):
            raise TypeError("first argument cell must be a cell")

        if len(cell1) != len(cell):
            return False

        mset1 = Counter(seq)
        mset2 = Counter(cell1.elements)
        if mset1 != mset2:
            return False

        size = len(seq)
        deq1 = deque(cell1.elements)
        deq2 = deque(seq)
        for _ in range(size):
            deq2.rotate()
            if deq1 == deq2:
                return True
        return False

    def __str__(self) -> str:
        """Return string representation of regular cell.

        Returns
        -------
        str
            Returns string representation of regular cell.
        """
        return f"Nodes set:{self.elements}, boundary edges:{self.boundary}, attrs:{self._attributes}"



"""HyperEdge classes."""

from collections.abc import Collection, Hashable

from toponetx.classes.complex import Atom

__all__ = ["HyperEdge"]


class HyperEdge(Atom[frozenset[Hashable]]):
    """Class for a hyperedge (or a set-type cell).

    This class represents a set-type cell in a combinatorial complex, which is a set of
    nodes with optional attributes and a rank. The nodes in a hyperedge must be
    hashable and unique, and the hyperedge itself is immutable.

    Parameters
    ----------
    elements : iterable of hashables
        The nodes in the hyperedge.
    rank : int, optional
        The rank of the hyperedge. Default is None.
    **kwargs : additional attributes
        Additional attributes of the hyperedge, as keyword arguments.

    Examples
    --------
    >>> ac1 = tnx.HyperEdge((1, 2, 3))
    >>> ac2 = tnx.HyperEdge((1, 2, 4, 5))
    >>> ac3 = tnx.HyperEdge(("a", "b", "c"))
    >>> ac3 = tnx.HyperEdge(("a", "b", "c"), rank=10)
    """

    def __init__(self, elements: Collection, rank=None, **kwargs) -> None:
        for i in elements:
            if not isinstance(i, Hashable):
                raise TypeError("Every element of HyperEdge must be hashable.")

        super().__init__(frozenset(sorted(elements)), **kwargs)
        if len(elements) != len(self.elements):
            raise ValueError("A hyperedge cannot contain duplicate nodes.")

        self._rank = rank

    def __str__(self) -> str:
        """Return a string representation of the HyperEdge.

        Returns
        -------
        str
            A string representation of the HyperEdge.
        """
        return f"Nodes set: {tuple(self.elements)}, attrs: {self._attributes}"

    def __repr__(self):
        """Return a string representation of the HyperEdge.

        Returns
        -------
        str
            A string representation of the HyperEdge.
        """
        return f"Nodes set: {tuple(self.elements)}, rank: {self._rank}"

    def __hash__(self):
        """Return a hash representation of the HyperEdge using its elements and rank.

        Returns
        -------
        int
            A hashed representation of the HyperEdge.
        """
        return hash((self.elements, self._rank))

    def __eq__(self, other):
        """Return whether all attributes of HyperEdge objects are equal.

        Parameters
        ----------
        other : HyperEdge
            The other HyperEdge that needs to be compared to the current HyperEdge.

        Returns
        -------
        bool
            A boolean value if the HyperEdge objects are equal.
        """
        if isinstance(other, HyperEdge):
            return self.__dict__ == other.__dict__  # check all attributes
        return NotImplemented

    def __ne__(self, other):
        """Return whether any attributes of HyperEdge objects are not equal.

        Parameters
        ----------
        other : HyperEdge
            The other HyperEdge that needs to be compared to the current HyperEdge.

        Returns
        -------
        bool
            A boolean value if the HyperEdge objects are not equal.
        """
        x = self.__eq__(other)
        if x is not NotImplemented:
            return not x
        return NotImplemented

    @property
    def rank(self):
        """Rank of the HyperEdge.

        Returns
        -------
        int or None
            The rank of the HyperEdge if it is not None, None otherwise.
        """
        if self._rank is not None:
            return self._rank
        return None



"""Unit tests for the tutorials."""

import subprocess
import tempfile
from pathlib import Path

import pytest


def _exec_tutorial(path):
    """Execute the notebooks for testing in the given directory.

    Parameters
    ----------
    path : str
        Path to the tutorials directory.
    """
    with tempfile.NamedTemporaryFile(suffix=".ipynb") as tmp_file:
        args = [
            "jupyter",
            "nbconvert",
            "--to",
            "notebook",
            "--execute",
            "--ExecutePreprocessor.timeout=1000",
            "--ExecutePreprocessor.kernel_name=python3",
            "--output",
            tmp_file.name,
            path,
        ]
        subprocess.check_call(args)  # noqa: S603


TUTORIALS_DIR = "tutorials"
paths = sorted(Path(TUTORIALS_DIR).glob("*.ipynb"))


@pytest.mark.parametrize("path", paths)
def test_tutorial(path):
    """Test the tutorials.

    Parameters
    ----------
    path : str
        Path to the tutorials directory.
    """
    _exec_tutorial(path)



"""Test distance module."""

import pytest

from toponetx.algorithms.distance import cell_distance, distance
from toponetx.classes.cell import Cell
from toponetx.classes.cell_complex import CellComplex
from toponetx.classes.combinatorial_complex import CombinatorialComplex
from toponetx.classes.hyperedge import HyperEdge
from toponetx.classes.simplicial_complex import SimplicialComplex
from toponetx.exception import TopoNetXNoPath


class TestDistance:
    """Test distance module."""

    def test_distance(self):
        """Test for the distance method."""
        CC = CellComplex()

        CC.add_cell([2, 3, 4], rank=2)
        CC.add_cell([5, 6, 7], rank=2)

        assert distance(CC, 2, 3) == 1

    def test_distance_edgecases(self):
        """Test the distance method to check exceptions."""
        CCC = CombinatorialComplex()

        CCC.add_cell([2, 3, 4], rank=2)
        CCC.add_cell([5, 6, 7], rank=2)

        with pytest.raises(TopoNetXNoPath):
            distance(CCC, 2, 5)

        with pytest.raises(KeyError):
            distance(CCC, [2, 3, 4], [5, 6, 7])

        with pytest.raises(KeyError):
            distance(CCC, Cell([2, 3, 4]), Cell([5, 6, 7]))

        with pytest.raises(TypeError):
            distance(1, 2, 3)

        with pytest.raises(TypeError):
            distance(SimplicialComplex(), 2, 3)

    def test_cell_distance(self):
        """Test for the cell_distance method."""
        CC = CellComplex()

        CC.add_cell([2, 3, 4], rank=2)
        CC.add_cell([5, 6, 7], rank=2)
        CC.add_cell([2, 5], rank=1)

        assert cell_distance(CC, (2, 3, 4), (5, 6, 7)) == 2

    def test_cell_distance_edgecases(self):
        """Test for cell_distance with exceptions."""
        CCC = CombinatorialComplex()

        CCC.add_cell([2, 3, 4], rank=2)
        CCC.add_cell([5, 6, 7], rank=2)

        with pytest.raises(TopoNetXNoPath):
            cell_distance(CCC, (2, 3, 4), (5, 6, 7))

        with pytest.raises(KeyError):
            cell_distance(CCC, [3, 4, 5], [5, 6, 7])

        with pytest.raises(KeyError):
            cell_distance(CCC, HyperEdge([3, 4, 5]), HyperEdge([5, 6, 7]))

        with pytest.raises(TypeError):
            cell_distance(1, 2, 3)

        with pytest.raises(TypeError):
            cell_distance(SimplicialComplex(), 2, 3)



"""Various examples of named meshes represented as complexes."""

import zipfile
from pathlib import Path
from typing import Literal, overload

import numpy as np
import requests

from toponetx.classes.cell_complex import CellComplex
from toponetx.classes.simplicial_complex import SimplicialComplex
from toponetx.readwrite.atomlist import load_from_atomlist

__all__ = ["coseg", "shrec_16", "stanford_bunny"]

DIR = Path(__file__).parent
SHREC_DS_MAP = {
    "full": "https://github.com/pyt-team/topological-datasets/raw/aa3ec28b1166f7713e872011df1ff9e1136d92ae/resources/shrec16_full.zip",
    "small": "https://github.com/pyt-team/topological-datasets/raw/aa3ec28b1166f7713e872011df1ff9e1136d92ae/resources/shrec16_small.zip",
}

COSEG_DS_MAP = {
    "alien": "https://github.com/pyt-team/topological-datasets/raw/aa3ec28b1166f7713e872011df1ff9e1136d92ae/resources/coseg_alien.zip",
    "chair": "https://github.com/pyt-team/topological-datasets/raw/aa3ec28b1166f7713e872011df1ff9e1136d92ae/resources/coseg_chair.zip",
    "vase": "https://github.com/pyt-team/topological-datasets/raw/aa3ec28b1166f7713e872011df1ff9e1136d92ae/resources/coseg_vase.zip",
}


@overload
def stanford_bunny(
    complex_type: Literal["cell"],
) -> CellComplex:  # numpydoc ignore=GL08
    ...


@overload
def stanford_bunny(
    complex_type: Literal["simplicial"],
) -> SimplicialComplex:  # numpydoc ignore=GL08
    ...


def stanford_bunny(
    complex_type: Literal["cell", "simplicial"],
) -> CellComplex | SimplicialComplex:
    """Load the Stanford Bunny mesh as a complex.

    Parameters
    ----------
    complex_type : {'cell', 'simplicial'}
        The type of complex to load. Supported values are
        "simplicial complex" and "cell complex".
        The default is "simplicial complex".

    Returns
    -------
    SimplicialComplex or CellComplex
        The loaded complex of the specified type.

    Raises
    ------
    ValueError
        If complex_type is not one of the supported values.
    """
    dataset_file = DIR / "bunny.obj"

    if not dataset_file.exists():
        r = requests.get(
            "https://github.com/pyt-team/topological-datasets/raw/main/resources/bunny.obj",
            timeout=10,
        )
        with dataset_file.open("wb") as f:
            f.write(r.content)

    if complex_type == "simplicial":
        return SimplicialComplex.load_mesh(dataset_file)
    if complex_type == "cell":
        return CellComplex.load_mesh(dataset_file)

    raise ValueError("complex_type must be 'simplicial' or 'cell'")


def shrec_16(size: Literal["full", "small"] = "full"):
    """Load training/testing shrec 16 datasets.

    Parameters
    ----------
    size : {'full', 'small'}, default='full'
        Dataset size. Options are "full" or "small".

    Returns
    -------
    tuple of length 2 npz files
        The npz files store the training/testing complexes of shrec 16 dataset along
        with their nodes, edges and faces features.

    Notes
    -----
    Each npz file stores 5 keys:
    "complexes","node_feat","edge_feat", "face_feat" and mesh label".
    complex : stores the simplicial complex of the mesh
    node_feat : stores a 6 dim node feature vector: position and normal of the each node in the mesh
    edge_feat : stores a 10 dim edge feature vector: dihedral angle, edge span, 2 edge angle in the triangle, 6 edge ratios.
    face_feat : stores a 7-dimensional face feature vector: face area (1 feat), face normal (3 feat), face angles (3 feat)
    mesh label : stores the label of the mesh

    Raises
    ------
    RuntimeError
        If dataset is not found on in DIR.

    Examples
    --------
    >>> shrec_training, shrec_testing = shrec_16()
    >>> # training dataset
    >>> training_complexes = shrec_training["complexes"]
    >>> training_labels = shrec_training["label"]
    >>> training_node_feat = shrec_training["node_feat"]
    >>> training_edge_feat = shrec_training["edge_feat"]
    >>> training_face_feat = shrec_training["face_feat"]
    >>> # testing dataset
    >>> testing_complexes = shrec_testing["complexes"]
    >>> testing_labels = shrec_testing["label"]
    >>> testing_node_feat = shrec_testing["node_feat"]
    >>> testing_edge_feat = shrec_testing["edge_feat"]
    >>> testing_face_feat = shrec_testing["face_feat"]
    """
    if size not in SHREC_DS_MAP:
        raise ValueError(
            f"Parameter `size` must be one of {list(SHREC_DS_MAP.keys())}, got {size}."
        )
    url = SHREC_DS_MAP[size]

    cache_file = DIR / f"shrec_{size}.zip"
    if not cache_file.exists():
        r = requests.get(url, timeout=10)
        with cache_file.open("wb") as f:
            f.write(r.content)

    with zipfile.ZipFile(cache_file) as zip_ref:
        train_complexes = []
        test_complexes = []

        for file in zip_ref.namelist():
            if file.endswith(".atomlist"):
                i = int(file.split("/")[1].split(".")[0])
                SC = load_from_atomlist(zip_ref.open(file), complex_type="simplicial")
                if "train" in file:
                    train_complexes.append((i, SC))
                if "test" in file:
                    test_complexes.append((i, SC))

        # `zip_ref.namelist` does not sort the files in the expected order
        train_complexes = [x[1] for x in sorted(train_complexes)]
        test_complexes = [x[1] for x in sorted(test_complexes)]

        train_data = {
            "complexes": train_complexes,
            "label": np.load(zip_ref.open("train/labels.npy")),
            "node_feat": np.load(
                zip_ref.open("train/node_feat.npy"), allow_pickle=True
            ),
            "edge_feat": np.load(zip_ref.open("train/edge_feat.npy")),
            "face_feat": np.load(zip_ref.open("train/face_feat.npy")),
        }
        test_data = {
            "complexes": test_complexes,
            "label": np.load(zip_ref.open("test/labels.npy")),
            "node_feat": np.load(zip_ref.open("test/node_feat.npy"), allow_pickle=True),
            "edge_feat": np.load(zip_ref.open("test/edge_feat.npy")),
            "face_feat": np.load(zip_ref.open("test/face_feat.npy")),
        }

    return train_data, test_data


def coseg(data: Literal["alien", "vase", "chair"] = "alien"):
    """Load coseg mesh segmentation datasets.

    The coseg dataset was downloaded and processed from the repo:
    https://github.com/Ideas-Laboratory/shape-coseg-dataset

    Parameters
    ----------
    data : {'alien', 'vase', 'chair'}, default='alien'
        The name of the coseg dataset to be loaded. Options are 'alien', 'vase', or 'chair'.

    Returns
    -------
    npz file
        The npz files store the complexes of coseg segmentation dataset along
        with their nodes, edges, and faces features.

    Raises
    ------
    RuntimeError
        If the dataset is not found in DIR.

    Notes
    -----
    Each npz file stores 5 keys:
    "complexes", "label", "node_feat", "edge_feat", and "face_feat".
    complex : stores the simplicial complex of the mesh
    node_feat : stores a 6-dimensional node feature vector: position and normal of each node in the mesh
    edge_feat : stores a 10-dimensional edge feature vector: dihedral angle, edge span, 2 edge angles in the triangle, 6 edge ratios.
    face_feat : stores a 7-dimensional face feature vector: face area (1 feat), face normal (3 feat), face angles (3 feat)
    face_label : stores the label of mesh segmentation as a face label

    Examples
    --------
    >>> coseg_data = coseg("alien")
    >>> complexes = coseg_data["complexes"]
    >>> node_feat = coseg_data["node_feat"]
    >>> edge_feat = coseg_data["edge_feat"]
    >>> face_feat = coseg_data["face_feat"]
    >>> face_label = coseg_data["face_label"]
    """
    if data not in COSEG_DS_MAP:
        raise ValueError(f"data must be 'alien', 'vase', or 'chair' got {data}.")
    url = COSEG_DS_MAP[data]

    cache_file = DIR / f"coseg_{data}.zip"
    if not cache_file.exists():
        r = requests.get(url, timeout=10)
        with cache_file.open("wb") as f:
            f.write(r.content)

    with zipfile.ZipFile(cache_file) as zip_ref:
        SCs = [
            (
                int(file.split(".")[0]),
                load_from_atomlist(zip_ref.open(file), complex_type="simplicial"),
            )
            for file in zip_ref.namelist()
            if file.endswith(".atomlist")
        ]

        SCs = [x[1] for x in sorted(SCs)]

        return {
            "complexes": SCs,
            "face_label": np.load(zip_ref.open("face_label.npy"), allow_pickle=True),
            "node_feat": np.load(zip_ref.open("node_feat.npy"), allow_pickle=True),
            "edge_feat": np.load(zip_ref.open("edge_feat.npy"), allow_pickle=True),
            "face_feat": np.load(zip_ref.open("face_feat.npy"), allow_pickle=True),
        }



"""
Metric and geometry backends for exterior calculus on triangle meshes embedded in R^3.

Notes
-----
This module implements the *metric/geometry* ingredients needed by (Discrete) Exterior Calculus (DEC)
and closely related FEM discretizations on triangle meshes.

High-level picture:
- A triangle mesh can be viewed as a 2D simplicial complex K embedded in R^3.
- Cochains C^k(K) assign scalars to oriented k-simplices (0: vertices, 1: edges, 2: triangles).
- Topology enters via boundary/coboundary operators; geometry enters via the Hodge star.

Continuous Hodge star (motivation):
- On an oriented Riemannian manifold (M, g), the Hodge star is an isomorphism
  * : Omega^k(M) -> Omega^{n-k}(M), determined by the metric g and the orientation.
- It is characterized by the identity
  a wedge (*b) = <a, b>_g vol_g
  for differential forms a, b of the same degree.

Discrete (diagonal) Hodge star (DEC):
- In DEC, one approximates the Hodge star by a sparse operator
  *_k : C^k(K) -> C^{n-k}(K^*),
  mapping primal k-cochains to dual (n-k)-cochains on a dual complex K^*.
- A common choice is a diagonal approximation:
      *_k ~= diag( |dual_k| / |primal_k| )
  where |primal_k| denotes a primal k-simplex measure and |dual_k| is the corresponding dual-cell
  measure (constructed e.g. using barycentric or circumcentric/Voronoi duals).

Robust Option A behavior for Voronoi/circumcentric presets
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Circumcentric/Voronoi duals can produce negative dual measures on meshes with obtuse triangles.
To keep the API simple and robust, the geometry backend uses the following default behavior
for presets "voronoi" and "circumcentric":

- For k=0 (vertex dual areas), we use a "mixed Voronoi area" construction:
  - non-obtuse triangles use Voronoi (cotangent) contributions,
  - obtuse triangles use an area split (A/2 to the obtuse vertex, A/4 to the others).
  This yields non-negative dual areas for non-degenerate triangles.
- For k=1 (dual edge lengths), we fall back to the barycentric dual edge length to avoid
  negative cotangent contributions on obtuse triangles.
- For k=2, the standard area relationship is used.

This module provides:
- Geometry utilities for triangles embedded in R^3.
- A user-facing `MetricSpec` describing diagonal Hodge stars and anisotropic FEM tensors.
- Backends implementing diagonal Hodge stars:
  - `DiagonalHodgeStar` (geometry-free).
  - `TriangleMesh3DBackend` (geometry-based).

Conventions:
- The mesh is treated as a 2D complex embedded in R^3.
- The Hodge stars returned here are diagonal (fast, standard in many DEC pipelines).
"""

from collections import defaultdict
from collections.abc import Callable
from dataclasses import dataclass
from typing import Any, Literal, Protocol

import numpy as np
from scipy.sparse import coo_matrix, csr_matrix, diags

from toponetx.classes.simplicial_complex import SimplicialComplex

MetricPreset = Literal[
    "identity",
    "diagonal",
    "barycentric_lumped",
    "circumcentric",
    "voronoi",
    "euclidean",
]

MetricCallable = Callable[[int, np.ndarray, SimplicialComplex], np.ndarray]


def _sorted_edge(u: Any, v: Any) -> tuple[Any, Any]:
    """Return an edge tuple with deterministic ordering.

    Parameters
    ----------
    u : Any
        First vertex label.
    v : Any
        Second vertex label.

    Returns
    -------
    tuple[Any, Any]
        Ordered pair (min(u, v), max(u, v)) under Python ordering.
    """
    return (u, v) if u <= v else (v, u)


def _triangle_area_3d(p0: np.ndarray, p1: np.ndarray, p2: np.ndarray) -> float:
    """Compute the area of a triangle embedded in R^3.

    Parameters
    ----------
    p0 : np.ndarray
        First vertex position, shape (3,).
    p1 : np.ndarray
        Second vertex position, shape (3,).
    p2 : np.ndarray
        Third vertex position, shape (3,).

    Returns
    -------
    float
        Triangle area.

    Raises
    ------
    ValueError
        If any input does not represent a 3D vector.
    """
    p0a = np.asarray(p0, dtype=float).reshape(-1)
    p1a = np.asarray(p1, dtype=float).reshape(-1)
    p2a = np.asarray(p2, dtype=float).reshape(-1)
    if p0a.shape != (3,) or p1a.shape != (3,) or p2a.shape != (3,):
        raise ValueError(
            "Triangle vertices must be 3D vectors with shape (3,). "
            f"Got shapes p0={p0a.shape}, p1={p1a.shape}, p2={p2a.shape}."
        )
    n = np.cross(p1a - p0a, p2a - p0a)
    return 0.5 * float(np.linalg.norm(n))


def _circumcenter_3d(p0: np.ndarray, p1: np.ndarray, p2: np.ndarray) -> np.ndarray:
    """Compute the circumcenter of a triangle embedded in R^3.

    Parameters
    ----------
    p0 : np.ndarray
        First vertex position, shape (3,).
    p1 : np.ndarray
        Second vertex position, shape (3,).
    p2 : np.ndarray
        Third vertex position, shape (3,).

    Returns
    -------
    np.ndarray
        Circumcenter position, shape (3,).

    Raises
    ------
    ValueError
        If any input does not represent a 3D vector.
    """
    a = np.asarray(p0, dtype=float).reshape(-1)
    b = np.asarray(p1, dtype=float).reshape(-1)
    c = np.asarray(p2, dtype=float).reshape(-1)
    if a.shape != (3,) or b.shape != (3,) or c.shape != (3,):
        raise ValueError(
            "Triangle vertices must be 3D vectors with shape (3,). "
            f"Got shapes p0={a.shape}, p1={b.shape}, p2={c.shape}."
        )
    ab = b - a
    ac = c - a
    n = np.cross(ab, ac)
    nn = float(n @ n)
    if nn < 1e-28:
        return (a + b + c) / 3.0
    ab2 = float(ab @ ab)
    ac2 = float(ac @ ac)
    u = (ac2 * np.cross(n, ab) + ab2 * np.cross(ac, n)) / (2.0 * nn)
    return a + u


def _grad_barycentric_3d(
    p0: np.ndarray, p1: np.ndarray, p2: np.ndarray
) -> tuple[np.ndarray, np.ndarray, np.ndarray, float]:
    """Compute gradients of P1 basis functions on an R^3 triangle.

    Parameters
    ----------
    p0 : np.ndarray
        First vertex position, shape (3,).
    p1 : np.ndarray
        Second vertex position, shape (3,).
    p2 : np.ndarray
        Third vertex position, shape (3,).

    Returns
    -------
    tuple[np.ndarray, np.ndarray, np.ndarray, float]
        Gradients (g0, g1, g2) each shape (3,), and det_norm = ||(p1-p0)x(p2-p0)||.
        The triangle area is A = 0.5 * det_norm.

    Raises
    ------
    ValueError
        If any input does not represent a 3D vector.
    """
    p0a = np.asarray(p0, dtype=float).reshape(-1)
    p1a = np.asarray(p1, dtype=float).reshape(-1)
    p2a = np.asarray(p2, dtype=float).reshape(-1)
    if p0a.shape != (3,) or p1a.shape != (3,) or p2a.shape != (3,):
        raise ValueError(
            "Triangle vertices must be 3D vectors with shape (3,). "
            f"Got shapes p0={p0a.shape}, p1={p1a.shape}, p2={p2a.shape}."
        )
    e01 = p1a - p0a
    e02 = p2a - p0a
    n = np.cross(e01, e02)
    nn = float(n @ n)
    det_norm = float(np.linalg.norm(n))
    if nn < 1e-28:
        z = np.zeros(3, dtype=float)
        return z, z, z, det_norm
    g0 = np.cross(n, p2a - p1a) / nn
    g1 = np.cross(n, p0a - p2a) / nn
    g2 = np.cross(n, p1a - p0a) / nn
    return g0.astype(float), g1.astype(float), g2.astype(float), det_norm


@dataclass(frozen=True)
class MetricSpec:
    """Describe how to build diagonal Hodge stars and anisotropic FEM tensors.

    Notes
    -----
    In DEC and related discretizations, the metric is represented by the discrete Hodge star.
    This class stores a *specification* for how diagonal stars (and optional anisotropic tensors)
    should be built.

    A common diagonal DEC choice is:
        *_k = diag(w_k), where w_k ~ |dual_k| / |primal_k|.

    For anisotropic diffusion / Riemannian metrics on the surface, one can also provide a
    per-triangle SPD tensor field G (a 3x3 matrix per triangle in embedding coordinates) that
    defines a stiffness operator assembled from P1 basis gradients.

    Attributes
    ----------
    preset : {"identity", "diagonal", "barycentric_lumped", "circumcentric", "voronoi", "euclidean"}
        Preset name controlling the backend.
    diagonal_weights : dict[int, np.ndarray], optional
        Explicit diagonal entries for the Hodge star at each degree k.
        Each entry must have length equal to the number of k-simplices.
    primal_measures : dict[int, np.ndarray], optional
        Primal measures |primal_k| for generic diagonal DEC stars.
    dual_measures : dict[int, np.ndarray], optional
        Dual measures |dual_k| for generic diagonal DEC stars.
    tensors : np.ndarray, optional
        Per-triangle SPD tensors for anisotropic stiffness, shape (nT, 3, 3).
    fn : callable, optional
        Per-triangle tensor function fn(t, P, sc) -> (3, 3), where P has shape (3, 3).
    eps : float
        Numerical safeguard used for divisions/inversions.
    """

    preset: MetricPreset = "barycentric_lumped"
    diagonal_weights: dict[int, np.ndarray] | None = None
    primal_measures: dict[int, np.ndarray] | None = None
    dual_measures: dict[int, np.ndarray] | None = None
    tensors: np.ndarray | None = None
    fn: MetricCallable | None = None
    eps: float = 1e-12


class HodgeStarBackend(Protocol):
    """Protocol for Hodge star backends."""

    def star(
        self, sc: SimplicialComplex, k: int, *, inverse: bool = False
    ) -> csr_matrix:
        """Return *_k (or its inverse).

        Parameters
        ----------
        sc : SimplicialComplex
            Complex object.
        k : int
            Cochain degree.
        inverse : bool, default=False
            If True, return (*_k)^{-1}.

        Returns
        -------
        csr_matrix
            Sparse star matrix.
        """
        ...


def _as_1d_float(x: np.ndarray) -> np.ndarray:
    """Convert an array-like to a 1D float array.

    Parameters
    ----------
    x : np.ndarray
        Input array-like.

    Returns
    -------
    np.ndarray
        1D float array.

    Raises
    ------
    ValueError
        If the input cannot be viewed as a 1D array after reshape.
    """
    arr = np.asarray(x, dtype=float).reshape(-1)
    if arr.ndim != 1:
        raise ValueError(f"Expected a 1D array after reshape, got shape {arr.shape}.")
    return arr


@dataclass(frozen=True)
class DiagonalHodgeStar:
    """Geometry-free diagonal Hodge star backend.

    Notes
    -----
    This backend does not use vertex positions. It builds diagonal stars from:
    - the identity preset, or
    - user-supplied diagonal weights / primal+dual measures.

    Attributes
    ----------
    metric : MetricSpec
        Metric specification.
    """

    metric: MetricSpec

    def star(
        self, sc: SimplicialComplex, k: int, *, inverse: bool = False
    ) -> csr_matrix:
        """Return a diagonal Hodge star matrix.

        Parameters
        ----------
        sc : SimplicialComplex
            Complex providing `skeleton(k)`.
        k : int
            Cochain degree.
        inverse : bool
            If True, return the inverse diagonal star.

        Returns
        -------
        csr_matrix
            Diagonal Hodge star matrix.
        """
        n_k = len(list(sc.skeleton(k)))
        eps = float(self.metric.eps)
        if self.metric.preset == "identity":
            diag = np.ones(n_k, dtype=float)
            return diags(diag, 0, format="csr")
        if self.metric.preset != "diagonal":
            raise ValueError(
                f"DiagonalHodgeStar requires preset='diagonal', got {self.metric.preset!r}."
            )
        if (
            self.metric.diagonal_weights is not None
            and k in self.metric.diagonal_weights
        ):
            diag = _as_1d_float(self.metric.diagonal_weights[k])
            if diag.shape[0] != n_k:
                raise ValueError(
                    f"diagonal_weights[{k}] must have length {n_k}, got {diag.shape[0]}."
                )
        else:
            if self.metric.primal_measures is None or self.metric.dual_measures is None:
                raise ValueError(
                    "preset='diagonal' requires diagonal_weights or both primal_measures and dual_measures."
                )
            pk = _as_1d_float(self.metric.primal_measures[k])
            dk = _as_1d_float(self.metric.dual_measures[k])
            if pk.shape[0] != n_k or dk.shape[0] != n_k:
                raise ValueError(
                    f"Measures for k={k} must have length {n_k}; got primal {pk.shape[0]}, "
                    f"dual {dk.shape[0]}."
                )
            diag = dk / np.maximum(pk, eps)
        if inverse:
            diag = 1.0 / np.maximum(diag, eps)
        return diags(diag, 0, format="csr")


@dataclass
class TriangleMesh3DBackend:
    """Geometry backend for triangle meshes embedded in R^3.

    Notes
    -----
    This backend precomputes mesh-aligned arrays (vertices, edges, triangles) from a
    `SimplicialComplex` with 3D vertex positions, and provides:
    - diagonal Hodge stars for k in {0, 1, 2} using barycentric-lumped or robust Voronoi-like
      constructions;
    - FEM operators on vertices (0-forms): mass, stiffness, and Laplacian-like operators;
    - anisotropic stiffness/Laplacian using per-triangle SPD tensors.

    Attributes
    ----------
    sc : SimplicialComplex
        Complex providing `skeleton(0/1/2)` and `get_node_attributes(pos_name)`.
    metric : MetricSpec
        Metric specification.
    pos_name : str
        Node attribute name for vertex positions in R^3.
    """

    sc: SimplicialComplex
    metric: MetricSpec
    pos_name: str = "position"

    def __post_init__(self) -> None:
        """Build the internal mesh cache."""
        self.eps = float(self.metric.eps)
        self._cache_ready = False
        self._build_cache()

    def _build_cache(self) -> None:
        """Build internal mesh cache."""
        tri = [tuple(s.elements) for s in self.sc.skeleton(2)]
        if not tri:
            raise ValueError("TriangleMesh3DBackend requires 2-simplices (triangles).")
        vlabels = [next(iter(s)) for s in self.sc.skeleton(0)]
        self._vid = {v: i for i, v in enumerate(vlabels)}
        pos = self.sc.get_node_attributes(self.pos_name)
        V = np.stack(
            [np.asarray(pos[v], dtype=float).reshape(-1) for v in vlabels], axis=0
        )
        if V.ndim != 2 or V.shape[1] != 3:
            raise ValueError(
                "Vertex positions must be 3D vectors (R^3) with shape (nV, 3). "
                f"Got shape {V.shape}."
            )
        self._V = V
        self._vlabels = vlabels
        edges = [tuple(s.elements) for s in self.sc.skeleton(1)]
        edges_sorted = [_sorted_edge(u, v) for (u, v) in edges]
        self._E = np.array(edges_sorted, dtype=object)
        self._T = np.array(tri, dtype=object)

        # Precompute per-vertex incident triangles (1-ring).
        nV = self._V.shape[0]
        self._incident: list[list[tuple[int, int, int]]] = [[] for _ in range(nV)]
        for t, (a, b, c) in enumerate(self._T):
            ia = self._vid[a]
            ib = self._vid[b]
            ic = self._vid[c]
            self._incident[ia].append((t, ib, ic))
            self._incident[ib].append((t, ia, ic))
            self._incident[ic].append((t, ia, ib))

        self._cache_ready = True

    def star(
        self, sc: SimplicialComplex, k: int, *, inverse: bool = False
    ) -> csr_matrix:
        """Return a diagonal Hodge star for triangle meshes.

        Parameters
        ----------
        sc : SimplicialComplex
            Complex (unused except for sizing consistency).
        k : int
            Cochain degree.
        inverse : bool, default=False
            If True, return (*_k)^{-1}.

        Returns
        -------
        csr_matrix
            Diagonal star matrix.
        """
        preset = self.metric.preset
        if preset == "identity":
            n = len(list(sc.skeleton(k)))
            return diags(np.ones(n, dtype=float), 0, format="csr")
        if preset == "barycentric_lumped":
            return self._star_barycentric_lumped(k, inverse=inverse)
        if preset in ("circumcentric", "voronoi"):
            # Robust Option A behavior (see module Notes).
            return self._star_voronoi_like_option_a(k, inverse=inverse)
        raise ValueError(
            f"TriangleMesh3DBackend does not support preset={preset!r} for stars."
        )

    def _star_barycentric_lumped(self, k: int, inverse: bool) -> csr_matrix:
        """Compute a barycentric-lumped diagonal star.

        Notes
        -----
        This constructs diagonal weights w_k = |dual_k| / |primal_k| using a barycentric dual.

        Parameters
        ----------
        k : int
            Cochain degree.
        inverse : bool, default=False
            If True, invert the diagonal star.

        Returns
        -------
        csr_matrix
            Diagonal star.
        """
        if k not in (0, 1, 2):
            raise ValueError(
                "barycentric_lumped supports only k in {0, 1, 2} for triangle meshes."
            )
        if k == 0:
            primal = np.ones(len(list(self.sc.skeleton(0))), dtype=float)
        elif k == 1:
            primal = self._primal_edge_lengths()
        else:
            primal = self._triangle_areas()
        dual = self._dual_measure_barycentric(k)
        w = dual / np.maximum(primal, self.eps)
        if inverse:
            w = 1.0 / np.maximum(w, self.eps)
        return diags(w, 0, format="csr")

    def _dual_measure_barycentric(self, k: int) -> np.ndarray:
        """Compute barycentric dual measures.

        Parameters
        ----------
        k : int
            Cochain degree.

        Returns
        -------
        np.ndarray
            Dual measures array for degree k.
        """
        A = self._triangle_areas()

        if k == 2:
            return A.copy()

        if k == 0:
            acc: dict[Any, float] = {}
            for (a, b, c), area in zip(self._T, A, strict=True):
                acc[a] = acc.get(a, 0.0) + area / 3.0
                acc[b] = acc.get(b, 0.0) + area / 3.0
                acc[c] = acc.get(c, 0.0) + area / 3.0
            return np.array([acc.get(v, 0.0) for v in self._vlabels], dtype=float)

        if k == 1:
            acc_e: dict[tuple[Any, Any], float] = {}

            # For barycentric dual on edges in 2D:
            # |e*| is approximated by summing, over incident triangles,
            # the segment length from the triangle barycenter to the edge midpoint.
            for (a, b, c), _area in zip(self._T, A, strict=True):
                ia = self._vid[a]
                ib = self._vid[b]
                ic = self._vid[c]

                pa = self._V[ia]
                pb = self._V[ib]
                pc = self._V[ic]

                bary = (pa + pb + pc) / 3.0  # triangle barycenter

                # edge (a, b)
                e = _sorted_edge(a, b)
                mid = (pa + pb) / 2.0
                acc_e[e] = acc_e.get(e, 0.0) + float(np.linalg.norm(bary - mid))

                # edge (a, c)
                e = _sorted_edge(a, c)
                mid = (pa + pc) / 2.0
                acc_e[e] = acc_e.get(e, 0.0) + float(np.linalg.norm(bary - mid))

                # edge (b, c)
                e = _sorted_edge(b, c)
                mid = (pb + pc) / 2.0
                acc_e[e] = acc_e.get(e, 0.0) + float(np.linalg.norm(bary - mid))

            edges = [tuple(e) for e in self._E.tolist()]
            return np.array([acc_e.get(tuple(e), 0.0) for e in edges], dtype=float)

        raise ValueError("k must be 0, 1, or 2.")

    def _star_voronoi_like_option_a(self, k: int, inverse: bool = False) -> csr_matrix:
        """Compute a robust Voronoi-like diagonal star (Option A).

        Notes
        -----
        This method is used for presets "voronoi" and "circumcentric" to avoid failures
        due to obtuse triangles. See the module Notes for the behavior.

        Parameters
        ----------
        k : int
            Cochain degree.
        inverse : bool, default=False
            If True, invert the diagonal star.

        Returns
        -------
        csr_matrix
            Diagonal star.

        Raises
        ------
        ValueError
            If k is not in {0, 1, 2}.
        """
        if k == 0:
            w = self._dual_area_vertices_voronoi_mixed()
            if inverse:
                w = 1.0 / np.maximum(w, self.eps)
            return diags(w, 0, format="csr")

        if k == 1:
            # Robust fallback: barycentric dual edge length, then star1 = |e*| / |e|.
            lstar = self._dual_measure_barycentric(1)
            le = self._primal_edge_lengths()
            w = lstar / np.maximum(le, self.eps)
            if inverse:
                w = 1.0 / np.maximum(w, self.eps)
            return diags(w, 0, format="csr")

        if k == 2:
            A = self._triangle_areas()
            w = 1.0 / np.maximum(A, self.eps)
            if inverse:
                w = A / np.maximum(1.0, self.eps)
            return diags(w, 0, format="csr")

        raise ValueError(
            "voronoi/circumcentric supports only k in {0, 1, 2} for triangle meshes."
        )

    def fem_mass_matrix_0(self) -> csr_matrix:
        """Return the consistent P1 FEM mass matrix on vertices.

        Returns
        -------
        csr_matrix
            Vertex-by-vertex mass matrix.
        """
        nV = self._V.shape[0]
        rows: list[int] = []
        cols: list[int] = []
        data: list[float] = []
        for a, b, c in self._T:
            ia, ib, ic = self._vid[a], self._vid[b], self._vid[c]
            A = _triangle_area_3d(self._V[ia], self._V[ib], self._V[ic])
            if A <= 1e-14:
                continue
            m = (A / 12.0) * np.array([[2, 1, 1], [1, 2, 1], [1, 1, 2]], dtype=float)
            idx = [ia, ib, ic]
            for i_local in range(3):
                for j_local in range(3):
                    rows.append(idx[i_local])
                    cols.append(idx[j_local])
                    data.append(float(m[i_local, j_local]))
        return coo_matrix((data, (rows, cols)), shape=(nV, nV)).tocsr()

    def cotan_stiffness_0(self) -> csr_matrix:
        """Return the surface FEM stiffness matrix on vertices.

        Returns
        -------
        csr_matrix
            Vertex-by-vertex stiffness matrix.
        """
        nV = self._V.shape[0]
        rows: list[int] = []
        cols: list[int] = []
        data: list[float] = []
        for a, b, c in self._T:
            ia, ib, ic = self._vid[a], self._vid[b], self._vid[c]
            p0, p1, p2 = self._V[ia], self._V[ib], self._V[ic]
            g0, g1, g2, det_norm = _grad_barycentric_3d(p0, p1, p2)
            A = 0.5 * det_norm
            if A <= 1e-14:
                continue
            grads = [g0, g1, g2]
            idx = [ia, ib, ic]
            for i_local in range(3):
                for j_local in range(3):
                    kij = A * float(grads[i_local] @ grads[j_local])
                    rows.append(idx[i_local])
                    cols.append(idx[j_local])
                    data.append(kij)
        return coo_matrix((data, (rows, cols)), shape=(nV, nV)).tocsr()

    def riemannian_stiffness_0(self) -> csr_matrix:
        """Return anisotropic surface stiffness matrix on vertices.

        Returns
        -------
        csr_matrix
            Vertex-by-vertex anisotropic stiffness matrix.
        """
        G = self._resolve_metric_tensors_per_triangle()
        nV = self._V.shape[0]
        rows: list[int] = []
        cols: list[int] = []
        data: list[float] = []
        for t, (a, b, c) in enumerate(self._T):
            ia, ib, ic = self._vid[a], self._vid[b], self._vid[c]
            p0, p1, p2 = self._V[ia], self._V[ib], self._V[ic]
            g0, g1, g2, det_norm = _grad_barycentric_3d(p0, p1, p2)
            A = 0.5 * det_norm
            if A <= 1e-14:
                continue
            GT = np.asarray(G[t], dtype=float)
            if GT.shape != (3, 3):
                raise ValueError(
                    f"Per-triangle tensor must have shape (3, 3), got {GT.shape}."
                )
            grads = [g0, g1, g2]
            idx = [ia, ib, ic]
            for i_local in range(3):
                for j_local in range(3):
                    kij = A * float(grads[i_local].T @ (GT @ grads[j_local]))
                    rows.append(idx[i_local])
                    cols.append(idx[j_local])
                    data.append(kij)
        return coo_matrix((data, (rows, cols)), shape=(nV, nV)).tocsr()

    def cotan_laplacian_0(self, *, lumped: bool = True) -> csr_matrix:
        """Return the cotangent Laplacian on vertices.

        Parameters
        ----------
        lumped : bool, default=True
            If True, return M_lumped^{-1} K. Otherwise return K.

        Returns
        -------
        csr_matrix
            Laplacian-like operator matrix.
        """
        K = self.cotan_stiffness_0()
        if not lumped:
            return K
        M0 = self.fem_mass_matrix_0()
        d = np.asarray(M0.sum(axis=1)).reshape(-1)
        Minv = diags(1.0 / np.maximum(d, self.eps), 0, format="csr")
        return Minv @ K

    def riemannian_laplacian_0(self, *, lumped: bool = True) -> csr_matrix:
        """Return anisotropic Laplacian on vertices.

        Parameters
        ----------
        lumped : bool, default=True
            If True, return M_lumped^{-1} K(G). Otherwise return K(G).

        Returns
        -------
        csr_matrix
            Laplacian-like operator matrix.
        """
        K = self.riemannian_stiffness_0()
        if not lumped:
            return K
        M0 = self.fem_mass_matrix_0()
        d = np.asarray(M0.sum(axis=1)).reshape(-1)
        Minv = diags(1.0 / np.maximum(d, self.eps), 0, format="csr")
        return Minv @ K

    def _triangle_areas(self) -> np.ndarray:
        """Compute triangle areas.

        Returns
        -------
        np.ndarray
            Areas of all triangles.
        """
        A = np.zeros((len(self._T),), dtype=float)
        for t, (a, b, c) in enumerate(self._T):
            ia, ib, ic = self._vid[a], self._vid[b], self._vid[c]
            A[t] = _triangle_area_3d(self._V[ia], self._V[ib], self._V[ic])
        return A

    def _primal_edge_lengths(self) -> np.ndarray:
        """Compute primal edge lengths.

        Returns
        -------
        np.ndarray
            Lengths of all edges.
        """
        le = np.zeros((len(self._E),), dtype=float)
        for i, (u, v) in enumerate(self._E):
            pu = self._V[self._vid[u]]
            pv = self._V[self._vid[v]]
            le[i] = float(np.linalg.norm(pu - pv))
        return le

    def _cot_from_squared_lengths(self, a2: float, b2: float, c2: float) -> float:
        """Compute cot(angle opposite side with squared length c2) from squared lengths.

        Parameters
        ----------
        a2 : float
            Squared length of one adjacent side.
        b2 : float
            Squared length of the other adjacent side.
        c2 : float
            Squared length of the opposite side.

        Returns
        -------
        float
            Cotangent of the angle opposite the side with squared length c2.
        """
        # cos(theta) = (a2 + b2 - c2) / (2ab)
        denom = 2.0 * np.sqrt(max(a2 * b2, 0.0)) + self.eps
        cos_t = (a2 + b2 - c2) / denom
        sin2 = max(1.0 - cos_t * cos_t, 0.0)
        sin_t = np.sqrt(max(sin2, self.eps))
        return float(cos_t / sin_t)

    def _dual_area_vertices_voronoi_mixed(self) -> np.ndarray:
        """Compute mixed Voronoi (circumcentric) dual area per vertex.

        Notes
        -----
        This is the standard "mixed area" construction used to avoid negative Voronoi areas
        on meshes with obtuse triangles:
        - If a triangle is non-obtuse, use Voronoi (cotangent) contributions.
        - If a triangle is obtuse, split its area: A/2 to the obtuse vertex, A/4 to the others.

        Returns
        -------
        np.ndarray
            Array of dual (mixed Voronoi) areas for each vertex, shape (n_vertices,).
        """
        nV = self._V.shape[0]
        Astar = np.zeros(nV, dtype=float)

        for a, b, c in self._T:
            ia, ib, ic = self._vid[a], self._vid[b], self._vid[c]
            pa, pb, pc = self._V[ia], self._V[ib], self._V[ic]

            # Squared edge lengths opposite vertices a, b, c.
            lab2 = float(np.dot(pa - pb, pa - pb))
            lbc2 = float(np.dot(pb - pc, pb - pc))
            lca2 = float(np.dot(pc - pa, pc - pa))

            # Angle obtuse test via squared lengths: angle at a is obtuse iff opposite edge bc is longest
            # and lbc2 > lab2 + lca2.
            obt_a = lbc2 > lab2 + lca2
            obt_b = lca2 > lab2 + lbc2
            obt_c = lab2 > lbc2 + lca2

            A = _triangle_area_3d(pa, pb, pc)
            if A <= 1e-14:
                continue

            if obt_a or obt_b or obt_c:
                # Mixed area split for obtuse triangles.
                if obt_a:
                    Astar[ia] += 0.5 * A
                    Astar[ib] += 0.25 * A
                    Astar[ic] += 0.25 * A
                elif obt_b:
                    Astar[ib] += 0.5 * A
                    Astar[ia] += 0.25 * A
                    Astar[ic] += 0.25 * A
                else:
                    Astar[ic] += 0.5 * A
                    Astar[ia] += 0.25 * A
                    Astar[ib] += 0.25 * A
                continue

            # Non-obtuse: Voronoi area formula using cotangents.
            # Contribution at vertex a: (|ab|^2 cot(gamma) + |ac|^2 cot(beta)) / 8
            # where beta is angle at b? careful:
            # - cot at b is opposite edge ac => uses lengths (ab, bc, ac)
            # - cot at c is opposite edge ab => uses lengths (bc, ca, ab)
            cot_b = self._cot_from_squared_lengths(
                lab2, lbc2, lca2
            )  # angle at b, opposite ca
            cot_c = self._cot_from_squared_lengths(
                lca2, lbc2, lab2
            )  # angle at c, opposite ab
            cot_a = self._cot_from_squared_lengths(
                lab2, lca2, lbc2
            )  # angle at a, opposite bc

            # Map squared lengths:
            # |ab|^2 = lab2, |ac|^2 = lca2, |bc|^2 = lbc2
            Astar[ia] += (lab2 * cot_c + lca2 * cot_b) / 8.0
            Astar[ib] += (lab2 * cot_c + lbc2 * cot_a) / 8.0
            Astar[ic] += (lca2 * cot_b + lbc2 * cot_a) / 8.0

        # Numerical guard: clip tiny negatives due to floating error.
        return np.maximum(Astar, 0.0)

    def _dual_area_vertices_circumcentric(self) -> np.ndarray:
        """Compute circumcentric/Voronoi dual areas per vertex using a cotangent formula.

        Notes
        -----
        This method is retained for reference and diagnostics. For robust default behavior on
        obtuse meshes, use `_dual_area_vertices_voronoi_mixed()` (Option A).

        Returns
        -------
        np.ndarray
            Array of dual (Voronoi) areas for each vertex, shape (n_vertices,).
        """
        nV = self._V.shape[0]
        Astar = np.zeros(nV, dtype=float)
        for v in range(nV):
            pv = self._V[v]
            for _, j, k in self._incident[v]:
                pj = self._V[j]
                pk = self._V[k]
                vj = pj - pv
                vk = pk - pv
                jk = pk - pj
                lj2 = float(np.dot(vj, vj))
                lk2 = float(np.dot(vk, vk))
                ljk2 = float(np.dot(jk, jk))

                denom_j = 2.0 * np.sqrt(lj2 * ljk2) + self.eps
                cos_j = (lj2 + ljk2 - lk2) / denom_j
                denom_k = 2.0 * np.sqrt(lk2 * ljk2) + self.eps
                cos_k = (lk2 + ljk2 - lj2) / denom_k

                sin_j_sq = max(1.0 - cos_j * cos_j, 0.0)
                sin_j = np.sqrt(max(sin_j_sq, self.eps))
                cot_j = cos_j / sin_j

                sin_k_sq = max(1.0 - cos_k * cos_k, 0.0)
                sin_k = np.sqrt(max(sin_k_sq, self.eps))
                cot_k = cos_k / sin_k

                Astar[v] += (lk2 * cot_j + lj2 * cot_k) / 8.0
        return Astar

    def _dual_length_edges_circumcentric(self) -> np.ndarray:
        """Compute circumcentric dual edge lengths using a cotangent formula.

        Notes
        -----
        This method is retained for reference and diagnostics. Under Option A, the default
        star for k=1 uses barycentric dual edge lengths to avoid negative cotangent sums.

        Returns
        -------
        np.ndarray
            Array of dual edge lengths, shape (n_edges,).
        """
        nE = len(self._E)
        lstar = np.zeros(nE, dtype=float)
        primal_len = self._primal_edge_lengths()

        edge_to_opp = defaultdict(list)
        for a, b, c in self._T:
            edge_to_opp[_sorted_edge(a, b)].append(c)
            edge_to_opp[_sorted_edge(a, c)].append(b)
            edge_to_opp[_sorted_edge(b, c)].append(a)

        for ei, (u, v) in enumerate(self._E):
            key = _sorted_edge(u, v)
            pu = self._V[self._vid[u]]
            pv = self._V[self._vid[v]]
            le = primal_len[ei]

            cot_sum = 0.0
            for opp_label in edge_to_opp[key]:
                opp_idx = self._vid[opp_label]
                po = self._V[opp_idx]

                ou = pu - po
                ov = pv - po
                lu2 = float(np.dot(ou, ou))
                lv2 = float(np.dot(ov, ov))
                uv2 = float(np.dot(pu - pv, pu - pv))

                denom = 2.0 * np.sqrt(lu2 * lv2) + self.eps
                cos_opp = (lu2 + lv2 - uv2) / denom

                sin_sq = max(1.0 - cos_opp * cos_opp, 0.0)
                sin_val = np.sqrt(max(sin_sq, self.eps))
                cot_opp = cos_opp / sin_val

                cot_sum += cot_opp

            lstar[ei] = (le / 2.0) * cot_sum

        return lstar

    def _resolve_metric_tensors_per_triangle(self) -> np.ndarray:
        """Resolve per-triangle SPD tensors.

        Returns
        -------
        np.ndarray
            Tensor array of shape (nT, 3, 3).

        Raises
        ------
        ValueError
            If `tensors` has the wrong shape or `fn` does not return a (3, 3) matrix.
        """
        nT = len(self._T)
        if self.metric.tensors is not None:
            G = np.asarray(self.metric.tensors, dtype=float)
            if G.shape != (nT, 3, 3):
                raise ValueError(
                    f"metric.tensors must have shape {(nT, 3, 3)}, got {G.shape}."
                )
            return G
        if self.metric.fn is not None:
            G = np.zeros((nT, 3, 3), dtype=float)
            for t, (a, b, c) in enumerate(self._T):
                P = np.stack(
                    [
                        self._V[self._vid[a]],
                        self._V[self._vid[b]],
                        self._V[self._vid[c]],
                    ],
                    axis=0,
                )
                GT = np.asarray(self.metric.fn(t, P, self.sc), dtype=float)
                if GT.shape != (3, 3):
                    raise ValueError("metric.fn must return a (3, 3) matrix.")
                G[t] = GT
            return G
        return np.repeat(np.eye(3, dtype=float)[None, :, :], nT, axis=0)



"""Test Cell class."""

import pytest

from toponetx.classes.simplex import Simplex


class TestSimplex:
    """Test the Simplex class."""

    def test_simplex_creation(self):
        """Test simplex creation."""
        s = Simplex((1,))
        assert len(s) == 1
        assert tuple(s) == (1,)
        assert s._attributes == {}

        with pytest.raises(ValueError):
            _ = Simplex([1, 2, 2])

        with pytest.raises(ValueError):
            _ = Simplex([1, 2], is_maximal=True)
        with pytest.raises(ValueError):
            _ = Simplex([1, 2], membership={})

    def test_simplex_sign(self):
        """Test simplex sign method."""
        s = Simplex([1])
        with pytest.raises(NotImplementedError):
            s.sign(face=1)

    def test_duplicate_nodes(self):
        """Test creation of simplex with duplicate nodes."""
        with pytest.raises(ValueError):
            Simplex((1, 1, 2))

    def test_contains(self):
        """Test the __contains__ method of the simplex."""
        s = Simplex((1, 2))
        assert 1 in s
        assert (1,) in s
        assert 3 not in s
        assert (1, 2) in s

    def test_le(self) -> None:
        """Test the __le__ method of the simplex."""
        s1 = Simplex([1, 2])
        s2 = Simplex([1, 2])
        s3 = Simplex([1, 2, 3])
        s4 = Simplex([4])

        assert s1 <= s2
        assert s1 <= s3
        assert not s3 <= s1
        assert s1 <= s4
        assert s3 <= s4

        with pytest.raises(TypeError):
            _ = s1 <= 1

    def test_representation(self):
        """Test the string representation of the simplex."""
        s = Simplex((1, 2, 3))
        assert repr(s), "Simplex((1, 2, 3))"
        assert str(s) == "Nodes set: (1, 2, 3), attrs: {}"

    def test__contains__(self):
        """Test the __contains__ method of the simplex."""
        s = Simplex([1, 2, 3])
        assert (1, 2) in s
        assert (1, 3) in s
        assert frozenset((1, 3)) in s
        assert (1, 3, 4, 5, 6, 7) not in s
        assert (1, 3, 4, 5, 6, 7) not in s

    def test_getting_and_setting_items(self):
        """Test getting and setting items in the simplex."""
        s = Simplex((1, 2, 3))
        s["weight"] = 1
        assert s["weight"] == 1

    def test_clone(self):
        """Test clone method."""
        s = Simplex((1, 2, 3))
        s1 = s.clone()

        assert 1 in s1
        assert 2 in s1
        assert 3 in s1
        assert (1, 2) in s1
        assert (2, 3) in s1
        assert (1, 3) in s1
        assert (1, 2, 3) in s1



"""Unit tests for the combinatorial complex class."""

import networkx as nx
import numpy as np
import pytest

from toponetx.classes.combinatorial_complex import CombinatorialComplex
from toponetx.classes.hyperedge import HyperEdge


class TestCombinatorialComplex:
    """Test CombinatorialComplex class."""

    def test_init_empty_ccc(self):
        """Test creation of an empty CCC."""
        CCC = CombinatorialComplex()
        assert len(CCC) == 0

    def test_init_from_lists(self):
        """Test creation of a CCC from a list of cells."""
        CCC = CombinatorialComplex([[1, 2, 3], [2, 3, 4]], ranks=2)
        assert len(CCC.cells) == 6
        assert (1, 2, 3) in CCC.cells
        assert (2, 3, 4) in CCC.cells

        with pytest.raises(ValueError):
            _ = CombinatorialComplex([[1, 2, 3], [2, 3, 4]], ranks=None)

        with pytest.raises(ValueError):
            _ = CombinatorialComplex([[1, 2, 3], [2, 3, 4]], ranks=[1])

    def test_init_from_abstract_cells(self):
        """Test creation of a CCC from abstract cells."""
        y1 = HyperEdge(elements=[1, 2], rank=1)
        y2 = HyperEdge(elements=[2, 4], rank=1)
        y3 = HyperEdge(elements=[3, 5], rank=1)
        y4 = HyperEdge(elements=[4, 5], rank=1)
        y5 = HyperEdge(elements=[5, 7], rank=1)
        y6 = HyperEdge(elements=[9, 8], rank=None)

        CCC = CombinatorialComplex(cells=[y1, y2, y3, y4, y5])

        assert y1 in CCC.cells
        assert y2 in CCC.cells
        assert y3 in CCC.cells
        assert y4 in CCC.cells
        assert y5 in CCC.cells

        with pytest.raises(ValueError):
            CCC = CombinatorialComplex(cells=[y1, y2, y3, y4, y5, y6])

    def test_init_from_networkx_graph(self):
        """Test creation of a CCC from a networkx graph."""
        G = nx.Graph()
        G.add_edge(0, 1)
        G.add_edge(0, 3)
        G.add_edge(0, 4)
        G.add_edge(1, 4)

        CCC = CombinatorialComplex(cells=G)

        assert (0, 1) in CCC.cells
        assert (0, 3) in CCC.cells
        assert (0, 4) in CCC.cells
        assert (1, 4) in CCC.cells
        assert (0, 5) not in CCC.cells

        G = nx.Graph()
        G.add_edge(5, 7)
        G.add_edge("a", "b")

        CCC.from_networkx_graph(G)

        assert "a" in CCC.cells

    def test_node_membership(self):
        """Test creation of _node_membership dictionary."""
        CCC = CombinatorialComplex()
        CCC.add_cell([1, 2, 3], rank=1)
        CCC.add_cell([1, 2, 3, 4], rank=2)

        assert 1 in CCC._node_membership
        assert 2 in CCC._node_membership
        assert 3 in CCC._node_membership
        assert 4 in CCC._node_membership
        assert frozenset({1, 2, 3, 4}) in CCC._node_membership[1]
        assert frozenset({1, 2, 3}) in CCC._node_membership[1]
        assert frozenset({1, 2, 3, 4}) in CCC._node_membership[2]
        assert frozenset({1, 2, 3}) in CCC._node_membership[2]
        assert frozenset({1, 2, 3, 4}) in CCC._node_membership[3]
        assert frozenset({1, 2, 3}) in CCC._node_membership[3]
        assert frozenset({1, 2, 3, 4}) in CCC._node_membership[4]

    def test_add_cell(self):
        """Test adding a cell to a CCC."""
        CCC = CombinatorialComplex()
        CCC.add_cell([1, 2, 3], rank=2)
        assert (1, 2, 3) in CCC.cells
        with pytest.raises(ValueError) as ex:
            CCC.add_cell(HyperEdge([1]), 3)
        assert len(str(ex.value)) > 0

        CCC_1 = CombinatorialComplex(graph_based=True)
        with pytest.raises(ValueError):
            CCC_1.add_cell(cell=[1, 2, 3], rank=1)
        with pytest.raises(TypeError):
            CCC_1.add_cell(cell=1, rank=1)

        with pytest.raises(ValueError):
            CCC.add_cell([1, 2], rank=-1)

        with pytest.raises(ValueError):
            CCC.add_cell(cell="[1,2]", rank=1)

        with pytest.raises(ValueError):
            CCC.add_cell(cell=1, rank=1)

        with pytest.raises(ValueError):
            CCC.add_cell(cell=[1, [2, 3]], rank=1)

        with pytest.raises(ValueError):
            CCC.add_cell(cell=[1, 2], rank=0)

    def test_add_cells_from(self):
        """Test adding multiple cells to a CCC."""
        CCC = CombinatorialComplex()
        CCC.add_cells_from([[2, 3, 4], [3, 4, 5]], ranks=2)

        assert (2, 3, 4) in CCC.cells
        assert (3, 4, 5) in CCC.cells

    def test_remove_cell(self):
        """Test removing a cell from a CCC."""
        CCC = CombinatorialComplex([[1, 2, 3], [2, 3, 4]], ranks=2)
        CCC.remove_cell([1, 2, 3])

        assert (1, 2, 3) not in CCC.cells
        assert (2, 3, 4) in CCC.cells

        with pytest.raises(KeyError):
            CCC.remove_cell([1, 2, 3])

        CCC.remove_cell(frozenset({1}))
        with pytest.raises(KeyError):
            CCC._complex_set.hyperedge_dict[0][frozenset({1})]

        CCC = CombinatorialComplex([(1, 2), (2, 3), (3, 4)], ranks=2)
        CCC.remove_cell(HyperEdge(elements=[1, 2], rank=1))

        assert (1, 2) not in CCC.cells

    def test_remove_cells(self):
        """Test removing multiple cells from a CCC."""
        CCC = CombinatorialComplex([[1, 2, 3], [2, 3, 4]], ranks=2)
        CCC.remove_cells([[1, 2, 3], [2, 3, 4]])

        assert (1, 2, 3) not in CCC.cells
        assert (2, 3, 4) not in CCC.cells

    def test_incidence_matrix(self):
        """Test generating an incidence matrix."""
        CCC = CombinatorialComplex([[1, 2, 3], [2, 3, 4]], ranks=2)

        B = CCC.incidence_matrix(rank=0, to_rank=2)
        assert B.shape == (4, 2)
        assert (B.T[0].todense() == [1, 1, 1, 0]).all()
        assert (B.T[1].todense() == [0, 1, 1, 1]).all()
        CCC = CombinatorialComplex()
        CCC.add_cell([1, 2], rank=1)
        CCC.add_cell([1, 2, 3, 4], rank=2)
        CCC.add_cell([1, 2], rank=1)
        CCC.add_cell([1, 3, 2], rank=1)
        CCC.add_cell([1, 2, 3, 4], rank=2)
        CCC.add_cell([2, 5], rank=1)
        CCC.add_cell([2, 6, 4], rank=2)
        B, _, _ = CCC.incidence_matrix(1, index=True)
        assert B[(frozenset({1, 2}))] == 0
        assert B[(frozenset({1, 2, 3}))] == 1
        assert B[(frozenset({2, 5}))] == 2

        with pytest.raises(ValueError):
            CCC.incidence_matrix(rank=2, to_rank=0)

    def test_incidence_matrix_to_rank_none(self):
        """Test generating an incidence matrix without setting the to_rank parameter."""
        CCC = CombinatorialComplex()
        CCC.add_cell([1, 2], rank=1)
        CCC.add_cell([1, 3], rank=1)
        CCC.add_cell([1, 2, 4, 3], rank=2)
        CCC.add_cell([2, 5], rank=1)
        CCC.add_cell([2, 6, 4], rank=2)
        B = CCC.incidence_matrix(0)
        assert B.shape == (6, 5)
        assert (
            B.todense()
            == [
                [1, 1, 0, 1, 0],
                [1, 0, 1, 1, 1],
                [0, 1, 0, 1, 0],
                [0, 0, 0, 1, 1],
                [0, 0, 1, 0, 0],
                [0, 0, 0, 0, 1],
            ]
        ).all()

    def test_adjacency_matrix(self):
        """Test generating an adjacency matrix."""
        CCC = CombinatorialComplex()
        CCC.add_cell([1, 2], rank=1)
        CCC.add_cell([1, 3], rank=1)
        CCC.add_cell([1, 2, 4, 3], rank=2)
        CCC.add_cell([2, 5], rank=1)
        CCC.add_cell([2, 6, 4], rank=2)
        A02 = CCC.adjacency_matrix(0, 2)
        assert A02.shape == (6, 6)

        with pytest.raises(ValueError):
            CCC.adjacency_matrix(2, 0)

    def test_coadjacency_matrix(self):
        """Test generating a coadjacency matrix."""
        CCC = CombinatorialComplex()
        CCC.add_cell([1, 2], rank=1)
        CCC.add_cell([1, 2], rank=1)
        CCC.add_cell([1, 3], rank=1)
        CCC.add_cell([1, 2, 4, 3], rank=2)
        CCC.add_cell([2, 5], rank=1)
        CCC.add_cell([2, 6, 4], rank=2)
        CA10 = CCC.coadjacency_matrix(1, 0)
        assert CA10.shape == (3, 3)
        assert (CA10.todense() == [[0, 1, 1], [1, 0, 0], [1, 0, 0]]).all()

        with pytest.raises(ValueError):
            CCC.coadjacency_matrix(0, 1)

    def test_dirac_operator_matrix(self):
        """Test dirac operator matrix."""
        CCC = CombinatorialComplex()
        CCC.add_cell([1, 2, 3, 4], rank=2)
        CCC.add_cell([1, 2], rank=1)
        CCC.add_cell([2, 3], rank=1)
        CCC.add_cell([1, 4], rank=1)
        CCC.add_cell([3, 4, 8], rank=2)
        m = CCC.dirac_operator_matrix()
        size = sum(CCC.shape)
        assert m.shape == (size, size)
        index, m = CCC.dirac_operator_matrix(index=True)
        assert frozenset({1, 2}) in index
        assert len(index) == size
        assert np.prod(m.todense() >= 0) == 1

    def test_clone(self):
        """Test the clone method of CombinatorialComplex."""
        CCC = CombinatorialComplex([[1, 2, 3], [2, 3, 4]], ranks=2)
        CCC2 = CCC.clone()
        assert len(CCC2.cells) == 6
        assert (1, 2, 3) in CCC2.cells
        assert (2, 3, 4) in CCC2.cells
        CCC2.remove_cell([1, 2, 3])
        assert len(CCC.cells) == 6

    def test_combinatorial_complex_init(self):
        """Test the init method of CombinatorialComplex class."""
        with pytest.raises(TypeError):
            CombinatorialComplex(cells=1)

    def test_incidence_matrix_to_rank_down_without_rank(self):
        """Test generating an incidence matrix by setting the down_rank parameter without mentioning rank."""
        CCC = CombinatorialComplex()
        CCC.add_cell([1, 2], rank=1)
        CCC.add_cell([1, 3], rank=1)
        CCC.add_cell([1, 2, 4, 3], rank=2)
        CCC.add_cell([2, 5], rank=1)
        CCC.add_cell([2, 6, 4], rank=2)
        B = CCC.incidence_matrix(2, incidence_type="down")
        assert B.shape == (9, 2)
        assert (
            B.todense()
            == [[1, 0], [1, 1], [1, 0], [1, 1], [0, 0], [0, 1], [1, 0], [1, 0], [0, 0]]
        ).all()

    def test_incidence_matrix_to_rank_with_wrong_incidence_type(self):
        """Test generating an incidence matrix by mentioning wrong rank."""
        CCC = CombinatorialComplex()
        CCC.add_cell([1, 2], rank=1)
        CCC.add_cell([1, 3], rank=1)
        CCC.add_cell([1, 2, 4, 3], rank=2)
        CCC.add_cell([2, 5], rank=1)
        CCC.add_cell([2, 6, 4], rank=2)
        with pytest.raises(ValueError) as exp:
            CCC.incidence_matrix(2, incidence_type="wrong")
        assert (
            str(exp.value) == "Invalid value for incidence_type. Must be 'up' or 'down'"
        )

    def test_incidence_matrix_with_equal_rank(self):
        """Test generating an incidence matrix by having equal rank."""
        CCC = CombinatorialComplex()
        CCC.add_cell([1, 2], rank=1)
        CCC.add_cell([1, 3], rank=1)
        CCC.add_cell([1, 2, 4, 3], rank=2)
        CCC.add_cell([2, 5], rank=1)
        CCC.add_cell([2, 6, 4], rank=2)
        with pytest.raises(ValueError) as exp:
            CCC.incidence_matrix(1, 1)
        assert (
            str(exp.value)
            == "incidence matrix can be computed for k!=r, got equal r and k."
        )

    def test_incidence_dict(self):
        """Test generating an incidence dictionary."""
        CCC = CombinatorialComplex()
        CCC.add_cell([1, 2], rank=1)
        CCC.add_cell([1, 3], rank=1)
        CCC.add_cell([1, 2, 4, 3], rank=2)
        CCC.add_cell([2, 5], rank=1)
        CCC.add_cell([2, 6, 4], rank=2)
        assert CCC.incidence_dict == {
            1: {
                frozenset({1, 2}): {"weight": 1},
                frozenset({1, 3}): {"weight": 1},
                frozenset({2, 5}): {"weight": 1},
            },
            0: {
                frozenset({1}): {"weight": 1},
                frozenset({2}): {"weight": 1},
                frozenset({3}): {"weight": 1},
                frozenset({4}): {"weight": 1},
                frozenset({5}): {"weight": 1},
                frozenset({6}): {"weight": 1},
            },
            2: {
                frozenset({1, 2, 3, 4}): {"weight": 1},
                frozenset({2, 4, 6}): {"weight": 1},
            },
        }

    def test_dim(self):
        """Test for the dimensionality of the CombinatorialComplex object."""
        CCC = CombinatorialComplex()
        CCC.add_cell([1, 2], rank=1)
        CCC.add_cell([1, 3], rank=1)
        CCC.add_cell([1, 2, 4, 3], rank=2)
        CCC.add_cell([2, 5], rank=1)
        CCC.add_cell([2, 6, 4], rank=3)
        assert CCC.dim == 3

    def test_repr(self):
        """Test the representation function of the CombinatorialComplex object by mentioning a name."""
        CCC = CombinatorialComplex()
        assert repr(CCC) == "CombinatorialComplex()"

    def test_contains(self):
        """Test whether the contains method works correctly."""
        CCC = CombinatorialComplex()
        CCC.add_cell([1, 2], rank=1)
        CCC.add_cell([1, 3], rank=1)
        CCC.add_cell([1, 2, 4, 3], rank=2)
        CCC.add_cell([2, 5], rank=1)
        CCC.add_cell([2, 6, 4], rank=2)

        assert 1 in CCC
        assert 3 in CCC
        assert 7 not in CCC

        assert (1, 2) in CCC
        assert (1, 4) not in CCC

        assert [(1)] in CCC.nodes
        assert [2] in CCC.nodes
        assert [3] in CCC.nodes
        assert [4] in CCC.nodes
        assert [5] in CCC.nodes
        assert [6] in CCC.nodes

    def test_set_item(self):
        """Test for set_item method of the CombinatorialComplex object."""
        CCC = CombinatorialComplex()
        CCC.add_cell([1], rank=0)
        CCC.add_cell([1, 2], rank=3)
        # Updating an node attribute present in the CombinatorialComplex object.
        CCC.__setitem__([1], {"weights": 1})
        # Setting a cell attribute present in the CombinatorialComplex object.
        CCC.__setitem__([1, 2], {"weights": 1})
        assert CCC._complex_set.hyperedge_dict[3][frozenset([1, 2])] == {"weights": 1}
        assert CCC.nodes[1]["weights"] == 1
        # Setting a cell attribute not present in the CombinatorialComplex object.
        with pytest.raises(KeyError):
            CCC.__setitem__([1, 2, 3], {"weights": 1})

    def test_degree(self):
        """Test for the degree function."""
        CCC = CombinatorialComplex()
        CCC.add_cell([1, 2], rank=1)
        CCC.add_cell([1, 3], rank=1)
        CCC.add_cell([1, 2, 4, 3], rank=2)
        CCC.add_cell([2, 5], rank=1)
        CCC.add_cell([2, 6, 4], rank=2)
        CCC.add_cell([1, 2], rank=2)
        assert CCC.degree(1) == 1
        with pytest.raises(ValueError) as exp:
            CCC.degree(1, -1)
        assert str(exp.value) == "Rank must be positive"
        assert CCC.degree(2, 2) == 3
        with pytest.raises(KeyError) as exp:
            node = 7
            assert CCC.degree(node, 2)
        assert str(exp.value) == f"'Node 7 not in {CCC.__shortstr__}.'"
        assert CCC.degree(1, rank=None) == 3

    def test_size(self):
        """Test for the size function."""
        CCC = CombinatorialComplex()
        CCC.add_cell([1, 2], rank=1)
        CCC.add_cell([1, 3], rank=1)
        CCC.add_cell([1, 2, 4, 3], rank=2)
        CCC.add_cell([2, 5], rank=1)
        CCC.add_cell([2, 6, 4], rank=2)
        CCC.add_cell([1, 2], rank=2)
        assert CCC.size(1) == 1
        with pytest.raises(ValueError) as exp:
            CCC.size(frozenset([1, 2, 3]))
        assert str(exp.value) == f"Input cell is not in cells of the {CCC.__shortstr__}"

    def test_num_nodes_and_cells(self):
        """Test for number of nodes and number of cells."""
        CCC = CombinatorialComplex()
        CCC.add_cell([1, 2], rank=1)
        CCC.add_cell([1, 3], rank=1)
        CCC.add_cell([1, 2, 4, 3], rank=2)
        CCC.add_cell([2, 5], rank=1)
        CCC.add_cell([2, 6, 4], rank=2)
        CCC.add_cell([1, 2], rank=2)
        y1 = HyperEdge([1, 2], rank=1)
        y2 = HyperEdge([1, 3], rank=1)
        assert CCC.number_of_nodes() == 6
        assert CCC.number_of_nodes([1, 2]) == 2
        assert CCC.number_of_cells() == 11
        assert CCC.number_of_cells([y1, y2]) == 2

    def test_order(self):
        """Test for the order function."""
        CCC = CombinatorialComplex()
        CCC.add_cell([1, 2, 3, 4, 5, 6, 7, 8], rank=1)
        assert CCC.order() == 8

    def test_remove_nodes(self):
        """Test for the remove_node and remove_nodes methods."""
        example = CombinatorialComplex()
        example.add_cell([1, 2], rank=1)
        example.add_cell([1, 3, 2], rank=1)
        example.add_cell([1, 2, 4, 3], rank=2)
        example.add_cell([2, 5], rank=1)
        example.add_cell([2, 6, 4], rank=2)
        example.remove_node(1)

        assert example._complex_set.hyperedge_dict == {
            1: {frozenset({2, 5}): {"weight": 1}},
            0: {
                frozenset({2}): {"weight": 1},
                frozenset({3}): {"weight": 1},
                frozenset({4}): {"weight": 1},
                frozenset({5}): {"weight": 1},
                frozenset({6}): {"weight": 1},
            },
            2: {frozenset({2, 4, 6}): {"weight": 1}},
        }
        with pytest.raises(KeyError) as exp:
            example.remove_nodes([1])
        assert str(exp.value) == f"'node 1 not in {example.__shortstr__}'"
        example.remove_nodes([2, 5])
        assert example._complex_set.hyperedge_dict == {
            0: {
                frozenset({3}): {"weight": 1},
                frozenset({4}): {"weight": 1},
                frozenset({6}): {"weight": 1},
            }
        }

        example.remove_nodes(HyperEdge([3]))
        assert example._complex_set.hyperedge_dict == {
            0: {frozenset({4}): {"weight": 1}, frozenset({6}): {"weight": 1}}
        }
        node = {4: 3}
        with pytest.raises(TypeError) as exp:
            example.remove_node(node)
        assert str(exp.value) == "node must be a HyperEdge or a hashable object"
        example = CombinatorialComplex()
        example.add_cell([1, 2], rank=1)
        example.add_cell([1, 3, 2], rank=1)
        example.add_cell([1, 2, 4, 3], rank=2)
        example.add_cell([2, 5], rank=1)
        example.add_cell([2, 6, 4], rank=2)
        example.remove_nodes([HyperEdge([1]), HyperEdge([2])])
        assert example._complex_set.hyperedge_dict == {
            0: {
                frozenset({3}): {"weight": 1},
                frozenset({4}): {"weight": 1},
                frozenset({5}): {"weight": 1},
                frozenset({6}): {"weight": 1},
            }
        }

    def test_set_cell_attributes(self):
        """Test for the set_cell_attributes method."""
        CCC = CombinatorialComplex()
        CCC.add_cell([1, 2], rank=1)
        CCC.add_cell([1, 3, 2], rank=1)
        CCC.add_cell([1, 2, 3, 4], rank=2)
        CCC.add_cell([2, 5], rank=1)
        CCC.add_cell([2, 6, 4], rank=2)
        d = {(1, 2, 3, 4): "red", (1, 2, 3): "blue"}
        CCC.set_cell_attributes(d, name="color")
        assert CCC.cells[(1, 2, 3, 4)]["color"] == "red"
        d = {(1, 2): {"attr1": "blue", "size": "large"}}
        CCC.set_cell_attributes(d)
        assert CCC.cells[(1, 2)]["attr1"] == "blue"

    def test_get_cell_attributes(self):
        """Test for get_cell_attributes method."""
        CCC = CombinatorialComplex()
        CCC.add_cell([1, 2], rank=1)
        CCC.add_cell([1, 3, 2], rank=1)
        CCC.add_cell([1, 2, 3, 4], rank=2)
        CCC.add_cell([2, 5], rank=1)
        CCC.add_cell([2, 6, 4], rank=2)
        d = {(1, 2, 3, 4): "red", (1, 2, 3): "blue"}
        CCC.set_cell_attributes(d, name="color")
        assert CCC.get_cell_attributes("color") == {
            frozenset({1, 2, 3}): "blue",
            frozenset({1, 2, 3, 4}): "red",
        }

    def test_set_node_attributes(self):
        """Test for the set and get nodes attributes method."""
        example1 = CombinatorialComplex()
        example1.add_cell([4, 6], rank=2)
        node = 4
        d = {node: {"color": "red", "attr2": 1}}
        example1.set_node_attributes(d)
        assert example1.nodes.nodes[frozenset({node})]["color"] == "red"
        assert example1.get_node_attributes("color") == {4: "red"}
        node = 6
        d = {node: "red"}
        example1.set_node_attributes(d, "color")
        # assert example1.nodes.
        assert example1.get_node_attributes("color") == {
            4: "red",
            6: "red",
        }

    def test_add_cells(self):
        """Test for the add_cells method."""
        CCC = CombinatorialComplex()
        CCC.add_cells_from([[1, 2], [1, 3, 2], [1, 2, 3, 4]], ranks=[1, 1, 2])
        assert CCC._complex_set.hyperedge_dict == {
            1: {frozenset({1, 2}): {"weight": 1}, frozenset({1, 2, 3}): {"weight": 1}},
            0: {
                frozenset({1}): {"weight": 1},
                frozenset({2}): {"weight": 1},
                frozenset({3}): {"weight": 1},
                frozenset({4}): {"weight": 1},
            },
            2: {frozenset({1, 2, 3, 4}): {"weight": 1}},
        }
        with pytest.raises(ValueError) as exp:
            CCC.add_cells_from([[1, 4]], ranks=[1, 1])
        assert str(exp.value) == "zip() argument 2 is longer than argument 1"
        CCC = CombinatorialComplex()
        CCC.add_cells_from(
            [
                HyperEdge([1, 2], rank=1),
                HyperEdge([1, 3, 2], rank=1),
                HyperEdge([1, 2, 3, 4], rank=2),
            ]
        )
        assert CCC._complex_set.hyperedge_dict == {
            1: {frozenset({1, 2}): {"weight": 1}, frozenset({1, 2, 3}): {"weight": 1}},
            0: {
                frozenset({1}): {"weight": 1},
                frozenset({2}): {"weight": 1},
                frozenset({3}): {"weight": 1},
                frozenset({4}): {"weight": 1},
            },
            2: {frozenset({1, 2, 3, 4}): {"weight": 1}},
        }
        CCC = CombinatorialComplex()
        with pytest.raises(ValueError) as exp:
            CCC.add_cells_from(
                [[1, 2], HyperEdge([1, 3, 2], rank=1), HyperEdge([1, 2, 3, 4], rank=2)]
            )
        assert (
            str(exp.value)
            == "input must be an HyperEdge [1, 2] object when rank is None"
        )
        with pytest.raises(ValueError) as exp:
            CCC.add_cells_from([HyperEdge([1, 3, 2], rank=1), HyperEdge([1, 2, 3, 4])])
        assert (
            str(exp.value)
            == "input HyperEdge Nodes set: (1, 2, 3, 4), attrs: {} has None rank"
        )

    def test_cell_node_adjacency_matrix(self):
        """Test for the cells adjacency matrix method."""
        CCC = CombinatorialComplex([[1, 2, 3], [2, 3, 4]], ranks=2)
        B = CCC.incidence_matrix(rank=0, to_rank=2)
        assert B.shape == (4, 2)
        assert (B.T[0].todense() == [1, 1, 1, 0]).all()
        assert (B.T[1].todense() == [0, 1, 1, 1]).all()
        CCC = CombinatorialComplex()
        CCC.add_cell([1, 2], rank=1)
        CCC.add_cell([1, 2, 3, 4], rank=2)
        CCC.add_cell([1, 2], rank=1)
        CCC.add_cell([1, 3, 2], rank=1)
        CCC.add_cell([1, 2, 3, 4], rank=2)
        CCC.add_cell([2, 5], rank=1)
        CCC.add_cell([2, 6, 4], rank=2)
        B, _, _ = CCC.incidence_matrix(1, index=True)
        assert B[(frozenset({1, 2}))] == 0
        assert B[(frozenset({1, 2, 3}))] == 1
        assert B[(frozenset({2, 5}))] == 2
        assert (
            CCC.node_to_all_cell_adjacnecy_matrix().todense()
            == [
                [0, 3, 2, 1, 0, 0],
                [3, 0, 2, 2, 1, 1],
                [2, 2, 0, 1, 0, 0],
                [1, 2, 1, 0, 0, 1],
                [0, 1, 0, 0, 0, 0],
                [0, 1, 0, 1, 0, 0],
            ]
        ).all()
        (
            CCC.all_cell_to_node_coadjacency_matrix().todense()
            == [
                [0, 2, 1, 2, 1],
                [2, 0, 1, 3, 1],
                [1, 1, 0, 1, 1],
                [2, 3, 1, 0, 2],
                [1, 1, 1, 2, 0],
            ]
        ).all()

    def test_singletons(self):
        """Test singletons of CCC."""
        CCC = CombinatorialComplex([[1, 2, 3], [2, 3, 4]], ranks=2)
        CCC.add_cell([9], rank=0)
        assert CCC.singletons() == [frozenset({9})]

    def test_remove_singletons(self):
        """Test remove_singletons of CCC."""
        CCC = CombinatorialComplex([[1, 2, 3], [2, 3, 4]], ranks=2)
        CCC.add_cell([9], rank=0)
        CCC = CCC.remove_singletons()
        assert 9 not in CCC.nodes

    def test_CCC_condition(self):
        """Test the _CCC_condition method of the CombinatorialComplex class."""
        CCC = CombinatorialComplex([[1, 2, 3], [2, 3, 4]], ranks=2)

        CCC._CCC_condition({1, 2}, rank=1)

        with pytest.raises(ValueError):
            CCC._CCC_condition({1, 2}, rank=3)



"""Methods to lift a graph to a simplicial complex."""

from itertools import combinations, takewhile

import networkx as nx

from toponetx.classes.simplicial_complex import SimplicialComplex

__all__ = [
    "graph_to_clique_complex",
    "graph_to_neighbor_complex",
    "weighted_graph_to_vietoris_rips_complex",
]


def graph_to_neighbor_complex(G: nx.Graph) -> SimplicialComplex:
    """Get the neighbor complex of a graph.

    Parameters
    ----------
    G : networkx.Graph
        Input graph.

    Returns
    -------
    toponetx.classes.SimplicialComplex
        The neighbor complex of the graph.

    Notes
    -----
    This type of simplicial complexes can have very large dimension (max degree of the
    graph) and it is a function of the distribution of the valency of the graph.
    """
    simplices = [[*list(G.neighbors(node)), node] for node in G]
    return SimplicialComplex(simplices)


def graph_to_clique_complex(
    G: nx.Graph, max_rank: int | None = None
) -> SimplicialComplex:
    """Get the clique complex of a graph.

    Parameters
    ----------
    G : networks.Graph
        Input graph.
    max_rank : int, optional
        The maximum rank of the simplices in the output clique complex.

    Returns
    -------
    SimplicialComplex
        The clique simplicial complex of rank `max_rank` of the graph G.
    """
    cliques = nx.enumerate_all_cliques(G)

    # `nx.enumerate_all_cliques` returns cliques in ascending order of size. Abort calling the generator once we reach
    # cliques larger than the requested max dimension.
    cliques = takewhile(
        lambda clique: max_rank is None or len(clique) <= max_rank + 1, cliques
    )

    SC = SimplicialComplex(cliques)

    # copy attributes of the input graph
    for node in G.nodes:
        SC[[node]].update(G.nodes[node])
    # if the resulting complex has dimension 1 or higher, copy the attributes of the edges
    if SC.dim >= 1:
        for edge in G.edges:
            SC[edge].update(G.edges[edge])
    SC.complex.update(G.graph)

    return SC


def weighted_graph_to_vietoris_rips_complex(
    G: nx.Graph, r: float, max_dim: int | None = None
):
    r"""Get the Vietoris-Rips complex of radius r of a weighted undirected graph.

    The Vietoris-Rips complex of radius `r` is the clique complex given by the cliques
    of `G` whose nodes have pairwise distances less or equal than `r`. All vertices are
    added to the Vietoris-Rips complex regardless of the radius introduced.

    If `G` is a clique weighted by a dissimilarity function d that satisfies
    \max_v d(v, v) <= \min d(u,v) for u != v, and r >= d(v, v) for all nodes v,
    then the Vietoris-Rips complex of radius `r` is the usual Vietoris-Rips abstract
    simplicial complex of radius `r` for point clouds with dissimilarities.

    Parameters
    ----------
    G : networkx.Graph
        Weighted undirected input graph. The weights of the edges must be in the attribute 'weight'.
    r : float
        The radius for the Vietoris-Rips simplicial complex computation.
    max_dim : int, optional
        The max dimension of the cliques in the output clique complex.

    Returns
    -------
    SimplicialComplex
        The Vietoris-Rips simplicial complex of dimension max_dim of the graph G.
    """

    def is_in_vr_complex(clique):  # numpydoc ignore=GL08
        edges = combinations(clique, 2)
        return all(G[u][v]["weight"] <= r for u, v in edges)

    all_cliques = nx.enumerate_all_cliques(G)
    possible_cliques = takewhile(
        lambda face: max_dim is None or len(face) <= max_dim, all_cliques
    )
    vr_cliques = filter(is_in_vr_complex, possible_cliques)
    return SimplicialComplex(list(vr_cliques))
